import copy
import csv
import datetime
import logging
import os
import re
import tempfile
from collections import defaultdict
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Optional

import boto3
import requests
from botocore.exceptions import ClientError, EndpointConnectionError

from gpuhunt._internal.models import QueryFilter, RawCatalogItem
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)
ec2_pricing_url = (
    "https://pricing.us-east-1.amazonaws.com/offers/v1.0/aws/AmazonEC2/current/index.csv"
)
disclaimer_rows_skip = 5
# https://aws.amazon.com/ec2/previous-generation/
previous_generation_families = [
    "a1.",
    "c1.",
    "c3.",
    "c4.",
    "g2.",
    "g3.",
    "g3s.",
    "i2.",
    "m1.",
    "m2.",
    "m3.",
    "p2.",
    "r3.",
    "r4.",
    "t1.",
    "cr1.",
    "hs1.",
]
pricing_filters = {
    "TermType": ["OnDemand"],
    "Tenancy": ["Shared"],
    "Operating System": ["Linux"],
    "CapacityStatus": ["Used"],
    "Unit": ["Hrs"],
    "Currency": ["USD"],
    "Pre Installed S/W": ["", "NA"],
    "MarketOption": ["OnDemand"],
}
describe_instances_limit = 100


class AWSProvider(AbstractProvider):
    """
    AWSProvider parses Bulk API index file for AmazonEC2 in all regions and fills missing GPU details

    Required IAM permissions:
    * `ec2:DescribeInstanceTypes`
    """

    NAME = "aws"

    def __init__(self, cache_path: Optional[str] = None):
        if cache_path:
            self.cache_path = cache_path
        else:
            self.temp_dir = tempfile.TemporaryDirectory()
            self.cache_path = self.temp_dir.name + "/index.csv"
        # todo aws creds
        self.preview_gpus = {
            "p4de.24xlarge": ("A100", 80.0),
        }

    def get(
        self, query_filter: Optional[QueryFilter] = None, balance_resources: bool = True
    ) -> list[RawCatalogItem]:
        if not os.path.exists(self.cache_path):
            logger.info("Downloading EC2 prices to %s", self.cache_path)
            with requests.get(ec2_pricing_url, stream=True, timeout=20) as r:
                r.raise_for_status()
                with open(self.cache_path, "wb") as f:
                    for chunk in r.iter_content(chunk_size=8192):
                        f.write(chunk)

        offers = []
        with open(self.cache_path, newline="") as f:
            for _ in range(disclaimer_rows_skip):
                f.readline()
            reader: Iterable[dict[str, str]] = csv.DictReader(f)
            for row in reader:
                if self.skip(row):
                    continue
                gpu_count = _parse_gpu_count(row["GPU"])
                if gpu_count is None:
                    continue
                offer = RawCatalogItem(
                    instance_name=row["Instance Type"],
                    location=row["Region Code"],
                    price=float(row["PricePerUnit"]),
                    cpu=int(row["vCPU"]),
                    memory=_parse_memory(row["Memory"]),
                    gpu_vendor=None,
                    gpu_count=gpu_count,
                    spot=False,
                    gpu_name=None,
                    gpu_memory=None,
                    disk_size=None,
                )
                offers.append(offer)
        self.fill_gpu_details(offers)
        offers = self.add_spots(offers)
        return sorted(offers, key=lambda i: i.price)

    def skip(self, row: dict[str, str]) -> bool:
        if any(row["Instance Type"].startswith(family) for family in previous_generation_families):
            return True
        for key, values in pricing_filters.items():
            if row[key] not in values:
                return True
        return False

    def fill_gpu_details(self, offers: list[RawCatalogItem]):
        regions = defaultdict(list)
        for offer in offers:
            if offer.gpu_count > 0 and offer.instance_name not in self.preview_gpus:
                regions[offer.location].append(offer.instance_name)

        gpus = copy.deepcopy(self.preview_gpus)
        while regions:
            region = max(regions, key=lambda r: len(regions[r]))
            instance_types = regions.pop(region)

            client = boto3.client("ec2", region_name=region)
            paginator = client.get_paginator("describe_instance_types")
            for offset in range(0, len(instance_types), describe_instances_limit):
                logger.info("Fetching GPU details for %s (offset=%s)", region, offset)
                pages = paginator.paginate(
                    InstanceTypes=instance_types[offset : offset + describe_instances_limit]
                )
                for page in pages:
                    for i in page["InstanceTypes"]:
                        if "GpuInfo" in i:
                            gpu = i["GpuInfo"]["Gpus"][0]
                            gpus[i["InstanceType"]] = (
                                gpu["Name"],
                                _get_gpu_memory_gib(gpu["Name"], gpu["MemoryInfo"]["SizeInMiB"]),
                            )

            regions = {
                region: left
                for region, names in regions.items()
                if (left := [i for i in names if i not in instance_types])
            }

        for offer in offers:
            if offer.gpu_count > 0:
                if offer.instance_name in gpus:
                    offer.gpu_name, offer.gpu_memory = gpus[offer.instance_name]
                else:
                    logger.warning(
                        "GPU info not available for instance type %s, skipping GPU details",
                        offer.instance_name,
                    )
                    offer.gpu_count = 0

    def _add_spots_worker(
        self, region: str, instance_types: set[str]
    ) -> dict[tuple[str, str], float]:
        spot_prices = dict()
        logger.info("Fetching spot prices for %s", region)
        try:
            client = boto3.client("ec2", region_name=region)  # todo creds
            pages = client.get_paginator("describe_spot_price_history").paginate(
                Filters=[
                    {
                        "Name": "product-description",
                        "Values": ["Linux/UNIX"],
                    }
                ],
                InstanceTypes=list(instance_types),
                StartTime=datetime.datetime.utcnow(),
            )

            instance_prices = defaultdict(list)
            for page in pages:
                for item in page["SpotPriceHistory"]:
                    instance_prices[item["InstanceType"]].append(float(item["SpotPrice"]))
            for (
                instance_type,
                zone_prices,
            ) in instance_prices.items():  # reduce zone prices to a single value
                spot_prices[(instance_type, region)] = min(zone_prices)
        except (ClientError, EndpointConnectionError):
            return {}
        return spot_prices

    def add_spots(self, offers: list[RawCatalogItem]) -> list[RawCatalogItem]:
        region_instances = defaultdict(set)
        for offer in offers:
            region_instances[offer.location].add(offer.instance_name)

        spot_prices = dict()
        with ThreadPoolExecutor(max_workers=8) as executor:
            future_to_region = {}
            for region, instance_types in region_instances.items():
                future = executor.submit(self._add_spots_worker, region, instance_types)
                future_to_region[future] = region
            for future in as_completed(future_to_region):
                spot_prices.update(future.result())

        spot_offers = []
        for offer in offers:
            if (price := spot_prices.get((offer.instance_name, offer.location))) is None:
                continue
            spot_offer = copy.deepcopy(offer)
            spot_offer.spot = True
            spot_offer.price = price
            spot_offers.append(spot_offer)
        return offers + spot_offers

    @classmethod
    def filter(cls, offers: list[RawCatalogItem]) -> list[RawCatalogItem]:
        return [
            i
            for i in offers
            if any(
                i.instance_name.startswith(family)
                for family in [
                    "m7i.",
                    "c7i.",
                    "r7i.",
                    "t3.",
                    "t2.small",
                    "c5.",
                    "m5.",
                    "p5.",
                    "p5e.",
                    "p4d.",
                    "p4de.",
                    "p3.",
                    "g6.",
                    "g6e.",
                    "gr6.",
                    "g5.",
                    "g4dn.",
                ]
            )
        ]


def _get_gpu_memory_gib(gpu_name: str, reported_memory_mib: int) -> float:
    """
    Fixes L4 memory size misreported by AWS API
    """

    if gpu_name != "L4":
        return reported_memory_mib / 1024

    if reported_memory_mib not in (22888, 91553, 183105):
        logger.warning(
            "The L4 memory size reported by AWS changed. "
            "Please check that it is now correct and remove the hardcoded size if it is."
        )
    return 24


def _parse_memory(s: str) -> float:
    r = re.match(r"^([0-9.]+) GiB$", s)
    return float(r.group(1))


def _parse_gpu_count(s: str) -> Optional[int]:
    if not s:
        return 0
    count = float(s)
    if count < 1:
        # AWS fractional GPUs not supported
        return None
    return int(count)



import csv
from operator import itemgetter
from pathlib import Path

import pytest


@pytest.fixture
def data_rows(catalog_dir: Path) -> list[dict]:
    with open(catalog_dir / "oci.csv") as f:
        return list(csv.DictReader(f))


@pytest.mark.parametrize("gpu", ["P100", "V100", "A10", "A100", ""])
def test_gpu_present(gpu: str, data_rows: list[dict]):
    assert gpu in map(itemgetter("gpu_name"), data_rows)


def test_on_demand_present(data_rows: list[dict]):
    assert "False" in map(itemgetter("spot"), data_rows)


def test_spot_present(data_rows: list[dict]):
    assert "True" in map(itemgetter("spot"), data_rows)


def test_spots_contain_flag(data_rows: list[dict]):
    for row in data_rows:
        assert (row["spot"] == "True") == ("oci-spot" in row["flags"]), row


@pytest.mark.parametrize("prefix", ["VM.Standard", "BM.Standard", "VM.GPU", "BM.GPU"])
def test_family_present(prefix: str, data_rows: list[dict]):
    assert any(name.startswith(prefix) for name in map(itemgetter("instance_name"), data_rows))


def test_quantity_decreases_as_query_complexity_increases(data_rows: list[dict]):
    zero_or_one_gpu = list(filter(lambda row: int(row["gpu_count"]) in (0, 1), data_rows))
    zero_gpu = list(filter(lambda row: int(row["gpu_count"]) == 0, data_rows))
    one_gpu = list(filter(lambda row: int(row["gpu_count"]) == 1, data_rows))

    assert len(data_rows) > len(zero_or_one_gpu)
    assert len(zero_or_one_gpu) > len(zero_gpu)
    assert len(zero_gpu) > len(one_gpu)



import csv
from operator import itemgetter
from pathlib import Path

import pytest


@pytest.fixture
def data_rows(catalog_dir: Path) -> list[dict]:
    with open(catalog_dir / "lambdalabs.csv") as f:
        return list(csv.DictReader(f))


@pytest.mark.parametrize("gpu", ["A10", "A100", "H100"])
def test_gpu_present(gpu: str, data_rows: list[dict]):
    assert gpu in map(itemgetter("gpu_name"), data_rows)


def test_on_demand_present(data_rows: list[dict]):
    assert "False" in map(itemgetter("spot"), data_rows)


def test_spot_not_present(data_rows: list[dict]):
    assert "True" not in map(itemgetter("spot"), data_rows)


def test_locations(data_rows: list[dict]):
    expected_locations = {
        "asia-northeast-1",
        "asia-northeast-2",
        "asia-south-1",
        "australia-east-1",
        "europe-central-1",
        "me-west-1",
        "us-east-1",
        "us-east-3",
        "us-midwest-1",
        "us-south-1",
        "us-south-2",
        "us-south-3",
        "us-west-1",
        "us-west-2",
        "us-west-3",
    }
    locations = set(map(itemgetter("location"), data_rows))
    missing = expected_locations - locations
    assert not missing



import enum
import json
from collections.abc import Container
from dataclasses import asdict, dataclass, field, fields
from typing import (
    ClassVar,
    Optional,
    Union,
)

from gpuhunt._internal.utils import empty_as_none

JSONType = Union[
    None,
    bool,
    int,
    float,
    str,
    list["JSONType"],
    "JSONObject",
]
JSONObject = dict[str, JSONType]


def bool_loader(x: Union[bool, str]) -> bool:
    if isinstance(x, bool):
        return x
    return x.lower() == "true"


class AMDArchitecture(enum.Enum):
    CDNA = "CDNA"
    CDNA2 = "CDNA2"
    CDNA3 = "CDNA3"

    @classmethod
    def cast(cls, value: Union["AMDArchitecture", str]) -> "AMDArchitecture":
        if isinstance(value, AMDArchitecture):
            return value
        return cls(value.upper())


class AcceleratorVendor(str, enum.Enum):
    NVIDIA = "nvidia"
    AMD = "amd"
    GOOGLE = "google"
    INTEL = "intel"
    TENSTORRENT = "tenstorrent"

    @classmethod
    def cast(cls, value: Union["AcceleratorVendor", str]) -> "AcceleratorVendor":
        if isinstance(value, AcceleratorVendor):
            return value
        return cls(value.lower())


class CPUArchitecture(str, enum.Enum):
    X86 = "x86"  # x86-64 extension support implied
    ARM = "arm"  # AArch64 (ARM64) execution state support implied

    @classmethod
    def cast(cls, value: Union["CPUArchitecture", str]) -> "CPUArchitecture":
        if isinstance(value, CPUArchitecture):
            return value
        return cls(value.lower())


@dataclass
class RawCatalogItem:
    """
    An item stored in the catalog.
    See `CatalogItem` for field descriptions.
    """

    instance_name: Optional[str]
    location: Optional[str]
    price: Optional[float]
    cpu: Optional[int]
    memory: Optional[float]
    gpu_count: Optional[int]
    gpu_name: Optional[str]
    gpu_memory: Optional[float]
    spot: Optional[bool]
    disk_size: Optional[float]
    gpu_vendor: Optional[str] = None
    flags: list[str] = field(default_factory=list)
    cpu_arch: Optional[str] = None
    provider_data: JSONObject = field(default_factory=dict)

    def __post_init__(self) -> None:
        self._process_gpu_vendor()
        self._process_cpu_arch()

    def _process_gpu_vendor(self) -> None:
        # This heuristic will be required indefinitely since we support historical catalogs.
        is_tpu = False
        gpu_name = self.gpu_name
        if gpu_name and gpu_name.startswith("tpu-"):
            is_tpu = True
            self.gpu_name = gpu_name[4:]
        gpu_vendor = self.gpu_vendor
        if gpu_vendor is None:
            if not self.gpu_count:
                # None or 0
                return
            if is_tpu:
                self.gpu_vendor = AcceleratorVendor.GOOGLE.value
            else:
                self.gpu_vendor = AcceleratorVendor.NVIDIA.value
        elif isinstance(gpu_vendor, AcceleratorVendor):
            self.gpu_vendor = gpu_vendor.value

    def _process_cpu_arch(self) -> None:
        # This heuristic will be required indefinitely since we support historical catalogs.
        cpu_arch = self.cpu_arch
        if cpu_arch is None:
            self.cpu_arch = CPUArchitecture.X86.value
        elif isinstance(cpu_arch, CPUArchitecture):
            self.cpu_arch = cpu_arch.value

    @staticmethod
    def from_dict(v: dict) -> "RawCatalogItem":
        return RawCatalogItem(
            instance_name=empty_as_none(v.get("instance_name")),
            location=empty_as_none(v.get("location")),
            price=empty_as_none(v.get("price"), loader=float),
            cpu_arch=empty_as_none(v.get("cpu_arch")),
            cpu=empty_as_none(v.get("cpu"), loader=int),
            memory=empty_as_none(v.get("memory"), loader=float),
            gpu_vendor=empty_as_none(v.get("gpu_vendor")),
            gpu_count=empty_as_none(v.get("gpu_count"), loader=int),
            gpu_name=empty_as_none(v.get("gpu_name")),
            gpu_memory=empty_as_none(v.get("gpu_memory"), loader=float),
            spot=empty_as_none(v.get("spot"), loader=bool_loader),
            disk_size=empty_as_none(v.get("disk_size"), loader=float),
            flags=v.get("flags", "").split(),
            provider_data=json.loads(v.get("provider_data", "{}")),
        )

    def dict(self) -> dict[str, Union[str, int, float, bool, None]]:
        return {
            **asdict(self),
            "flags": " ".join(self.flags),
            "provider_data": json.dumps(self.provider_data),
        }


@dataclass
class CatalogItem:
    """
    An item returned by `Catalog.query`.
    Attributes:
        instance_name: name of the instance
        location: region or zone
        price: $ per hour
        cpu_arch: CPU instruction set architecture
        cpu: number of CPUs
        memory: amount of RAM in GB
        gpu_vendor: GPU/accelerator vendor
        gpu_count: number of GPUs
        gpu_name: name of the GPU
        gpu_memory: amount of GPU VRAM in GB for each GPU
        spot: whether the instance is a spot instance
        provider: name of the provider
        disk_size: size of disk in GB
        flags: list of flags. If a catalog item breaks existing dstack versions,
            add a flag to hide the item from those versions. Newer dstack versions
            will have to request this flag explicitly to get the catalog item.
            If you are adding a new provider, leave the flags empty.
            Flag names should be in kebab-case.
        provider_data: dict with provider-specific properties.
            Prefer defining a TypedDict within provider implementation.
    """

    instance_name: str
    location: str
    price: float
    cpu: int
    memory: float
    gpu_count: int
    gpu_name: Optional[str]
    gpu_memory: Optional[float]
    spot: bool
    disk_size: Optional[float]
    provider: str
    gpu_vendor: Optional[AcceleratorVendor] = None
    flags: list[str] = field(default_factory=list)
    cpu_arch: Optional[CPUArchitecture] = None
    provider_data: JSONObject = field(default_factory=dict)

    def __post_init__(self) -> None:
        self._process_gpu_vendor()
        self._process_cpu_arch()

    def _process_gpu_vendor(self) -> None:
        # This heuristic is only required until we update all providers to always set the vendor.
        gpu_vendor = self.gpu_vendor
        if gpu_vendor is None:
            if not self.gpu_count:
                # None or 0
                return
            # GCPProvider already sets gpu_vendor, and all other providers only support Nvidia
            self.gpu_vendor = AcceleratorVendor.NVIDIA
        else:
            # This cast to the enum is always required since RawCatalogItem.gpu_vendor
            # is a string field (for (de)serialization purposes).
            self.gpu_vendor = AcceleratorVendor.cast(gpu_vendor)

    def _process_cpu_arch(self) -> None:
        # This heuristic is only required until we update all providers to always set the arch.
        cpu_arch = self.cpu_arch
        if cpu_arch is None:
            self.cpu_arch = CPUArchitecture.X86
        else:
            self.cpu_arch = CPUArchitecture.cast(cpu_arch)

    @staticmethod
    def from_dict(v: dict, *, provider: Optional[str] = None) -> "CatalogItem":
        return CatalogItem(provider=provider, **asdict(RawCatalogItem.from_dict(v)))


@dataclass
class QueryFilter:
    """
    Attributes:
        provider: name of the provider to filter by. If not specified, all providers will be used
        cpu_arch: CPU architecture. If not specified, all architectures will be used
        min_cpu: minimum number of CPUs
        max_cpu: maximum number of CPUs
        min_memory: minimum amount of RAM in GB
        max_memory: maximum amount of RAM in GB
        min_gpu_count: minimum number of GPUs
        max_gpu_count: maximum number of GPUs
        gpu_vendor: accelerator vendor to filter by. If not specified, all vendors will be used
        gpu_name: name of the GPU to filter by. If not specified, all GPUs will be used
        min_gpu_memory: minimum amount of GPU VRAM in GB for each GPU
        max_gpu_memory: maximum amount of GPU VRAM in GB for each GPU
        min_total_gpu_memory: minimum amount of GPU VRAM in GB for all GPUs combined
        max_total_gpu_memory: maximum amount of GPU VRAM in GB for all GPUs combined
        min_disk_size: minimum disk size in GB
        max_disk_size: maximum disk size in GB
        min_price: minimum price per hour in USD
        max_price: maximum price per hour in USD
        min_compute_capability: minimum compute capability of the GPU
        max_compute_capability: maximum compute capability of the GPU
        spot: if `False`, only ondemand offers will be returned. If `True`, only spot offers will be returned
        allowed_flags: only offers with all flags allowed will be returned. `None` allows all flags
    """

    provider: Optional[list[str]] = None  # strings can have mixed case
    cpu_arch: Optional[CPUArchitecture] = None
    min_cpu: Optional[int] = None
    max_cpu: Optional[int] = None
    min_memory: Optional[float] = None
    max_memory: Optional[float] = None
    min_gpu_count: Optional[int] = None
    max_gpu_count: Optional[int] = None
    gpu_vendor: Optional[AcceleratorVendor] = None
    gpu_name: Optional[list[str]] = None  # strings can have mixed case
    min_gpu_memory: Optional[float] = None
    max_gpu_memory: Optional[float] = None
    min_total_gpu_memory: Optional[float] = None
    max_total_gpu_memory: Optional[float] = None
    min_disk_size: Optional[int] = None
    max_disk_size: Optional[int] = None
    min_price: Optional[float] = None
    max_price: Optional[float] = None
    min_compute_capability: Optional[tuple[int, int]] = None
    max_compute_capability: Optional[tuple[int, int]] = None
    spot: Optional[bool] = None
    allowed_flags: Optional[Container[str]] = None

    def __repr__(self) -> str:
        """
        >>> QueryFilter()
        QueryFilter()
        >>> QueryFilter(min_cpu=4)
        QueryFilter(min_cpu=4)
        >>> QueryFilter(max_price=1.2, min_cpu=4)
        QueryFilter(min_cpu=4, max_price=1.2)
        """
        kv = ", ".join(
            f"{f.name}={value}"
            for f in fields(self)
            if (value := getattr(self, f.name)) is not None
        )
        return f"QueryFilter({kv})"


@dataclass
class AcceleratorInfo:
    vendor: ClassVar[AcceleratorVendor]
    name: str
    memory: int


@dataclass
class NvidiaGPUInfo(AcceleratorInfo):
    vendor = AcceleratorVendor.NVIDIA
    compute_capability: tuple[int, int]


@dataclass
class AMDGPUInfo(AcceleratorInfo):
    vendor = AcceleratorVendor.AMD
    architecture: AMDArchitecture
    device_ids: tuple[int, ...]


@dataclass
class TPUInfo(AcceleratorInfo):
    vendor = AcceleratorVendor.GOOGLE


@dataclass
class IntelAcceleratorInfo(AcceleratorInfo):
    vendor = AcceleratorVendor.INTEL


@dataclass
class TenstorrentAcceleratorInfo(AcceleratorInfo):
    vendor = AcceleratorVendor.TENSTORRENT



import os

import pytest

from gpuhunt.providers.hotaisle import HotAisleProvider


@pytest.fixture
def provider():
    api_key = os.environ.get("HOTAISLE_API_KEY")
    team_handle = os.environ.get("HOTAISLE_TEAM_HANDLE")
    return HotAisleProvider(api_key=api_key, team_handle=team_handle)


@pytest.fixture
def offers(provider):
    """Fixture that provides the list of offers from HotAisle."""
    return provider.get()


def test_positive_prices(offers):
    """Test that all offers have positive prices."""
    assert all(offer.price > 0 for offer in offers)



import argparse
import logging
import os

import gpuhunt._internal.storage as storage
from gpuhunt._internal.utils import configure_logging


def main():
    parser = argparse.ArgumentParser(prog="python3 -m gpuhunt")
    parser.add_argument(
        "provider",
        choices=[
            "aws",
            "azure",
            "cloudrift",
            "cudo",
            "datacrunch",
            "digitalocean",
            "gcp",
            "hotaisle",
            "lambdalabs",
            "nebius",
            "oci",
            "runpod",
            "tensordock",
            "vastai",
            "vultr",
        ],
    )
    parser.add_argument("--output", required=True)
    parser.add_argument("--no-filter", action="store_true")
    args = parser.parse_args()
    configure_logging()

    if args.provider == "aws":
        from gpuhunt.providers.aws import AWSProvider

        provider = AWSProvider(os.getenv("AWS_CACHE_PATH"))
    elif args.provider == "azure":
        from gpuhunt.providers.azure import AzureProvider

        provider = AzureProvider(os.getenv("AZURE_SUBSCRIPTION_ID"))
    elif args.provider == "cudo":
        from gpuhunt.providers.cudo import CudoProvider

        provider = CudoProvider()
    elif args.provider == "cloudrift":
        from gpuhunt.providers.cloudrift import CloudRiftProvider

        provider = CloudRiftProvider()
    elif args.provider == "datacrunch":
        from gpuhunt.providers.datacrunch import DataCrunchProvider

        provider = DataCrunchProvider(
            os.getenv("DATACRUNCH_CLIENT_ID"), os.getenv("DATACRUNCH_CLIENT_SECRET")
        )
    elif args.provider == "digitalocean":
        from gpuhunt.providers.digitalocean import DigitalOceanProvider

        provider = DigitalOceanProvider(
            api_key=os.getenv("DIGITAL_OCEAN_API_KEY"), api_url=os.getenv("DIGITAL_OCEAN_API_URL")
        )
    elif args.provider == "gcp":
        from gpuhunt.providers.gcp import GCPProvider

        provider = GCPProvider(os.getenv("GCP_PROJECT_ID"))
    elif args.provider == "hotaisle":
        from gpuhunt.providers.hotaisle import HotAisleProvider

        provider = HotAisleProvider(
            api_key=os.getenv("HOTAISLE_API_KEY"), team_handle=os.getenv("HOTAISLE_TEAM_HANDLE")
        )
    elif args.provider == "lambdalabs":
        from gpuhunt.providers.lambdalabs import LambdaLabsProvider

        provider = LambdaLabsProvider(os.getenv("LAMBDALABS_TOKEN"))
    elif args.provider == "nebius":
        from nebius.base.service_account.pk_file import Reader as PKReader

        from gpuhunt.providers.nebius import NebiusProvider

        provider = NebiusProvider(
            credentials=PKReader(
                filename=os.getenv("NEBIUS_PRIVATE_KEY_FILE"),
                public_key_id=os.getenv("NEBIUS_PUBLIC_KEY_ID"),
                service_account_id=os.getenv("NEBIUS_SERVICE_ACCOUNT_ID"),
            ),
        )
    elif args.provider == "oci":
        from gpuhunt.providers.oci import OCICredentials, OCIProvider

        provider = OCIProvider(
            OCICredentials(
                user=os.getenv("OCI_CLI_USER"),
                key_content=os.getenv("OCI_CLI_KEY_CONTENT"),
                fingerprint=os.getenv("OCI_CLI_FINGERPRINT"),
                tenancy=os.getenv("OCI_CLI_TENANCY"),
                region=os.getenv("OCI_CLI_REGION"),
            )
        )
    elif args.provider == "runpod":
        from gpuhunt.providers.runpod import RunpodProvider

        provider = RunpodProvider()
    elif args.provider == "tensordock":
        from gpuhunt.providers.tensordock import TensorDockProvider

        provider = TensorDockProvider()
    elif args.provider == "vastai":
        from gpuhunt.providers.vastai import VastAIProvider

        provider = VastAIProvider()
    elif args.provider == "vultr":
        from gpuhunt.providers.vultr import VultrProvider

        provider = VultrProvider()
    else:
        exit(f"Unknown provider {args.provider}")

    logging.info("Fetching offers for %s", args.provider)
    offers = provider.get()
    if not args.no_filter:
        offers = provider.filter(offers)
    storage.dump(offers, args.output)


if __name__ == "__main__":
    main()



import logging
import os
from typing import Optional

import requests
from requests import Response

from gpuhunt._internal.constraints import find_accelerators
from gpuhunt._internal.models import AcceleratorVendor, QueryFilter, RawCatalogItem
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)

API_URL = "https://admin.hotaisle.app/api"


class HotAisleProvider(AbstractProvider):
    NAME = "hotaisle"

    def __init__(self, api_key: Optional[str] = None, team_handle: Optional[str] = None):
        """Hotaisle requries an API key and team handle to access the API."""
        self.api_key = api_key or os.getenv("HOTAISLE_API_KEY")
        self.team_handle = team_handle or os.getenv("HOTAISLE_TEAM_HANDLE")

        if not self.api_key:
            raise ValueError("Set the HOTAISLE_API_KEY environment variable.")
        if not self.team_handle:
            raise ValueError("Set the HOTAISLE_TEAM_HANDLE environment variable.")

    def get(
        self, query_filter: Optional[QueryFilter] = None, balance_resources: bool = True
    ) -> list[RawCatalogItem]:
        offers = self.fetch_offers()
        return sorted(offers, key=lambda i: i.price)

    def fetch_offers(self) -> list[RawCatalogItem]:
        """Fetch available virtual machines from HotAisle API.
        See API documentation(https://admin.hotaisle.app/api/docs)
        for details."""
        url = f"/teams/{self.team_handle}/virtual_machines/available/"
        response = self._make_request("GET", url)
        return convert_response_to_raw_catalog_items(response)

    def _make_request(self, method: str, url: str) -> Response:
        full_url = f"{API_URL}{url}"
        headers = {
            "accept": "application/json",
            "Authorization": f"Token {self.api_key}",
        }

        response = requests.request(method=method, url=full_url, headers=headers, timeout=30)
        response.raise_for_status()
        return response


def get_gpu_memory(gpu_name: str) -> Optional[float]:
    if accelerators := find_accelerators(names=[gpu_name], vendors=[AcceleratorVendor.AMD]):
        return float(accelerators[0].memory)
    logger.warning(f"Unknown AMD GPU {gpu_name}")
    return None


def convert_response_to_raw_catalog_items(response: Response) -> list[RawCatalogItem]:
    data = response.json()
    offers = []
    for item in data:
        price_in_cents = item["OnDemandPrice"]
        price = float(price_in_cents) / 100
        specs = item["Specs"]
        cpu_cores = specs["cpu_cores"]
        ram_capacity_bytes = specs["ram_capacity"]
        memory_gb = ram_capacity_bytes / (1024**3)
        disk_capacity_bytes = specs["disk_capacity"]
        disk_gb = disk_capacity_bytes / (1024**3)
        cpus = specs["cpus"]
        cpu_model = cpus["model"]
        gpus = specs["gpus"]
        gpu = gpus[0]
        gpu_count = gpu["count"]
        gpu_name = gpu["model"]
        gpu_vendor = AcceleratorVendor.AMD.value  # All GPUs are AMD with HotAisle.
        gpu_memory = get_gpu_memory(gpu_name)

        # Create instance name: cpu_model-cores-ram-gpucount-gpu
        instance_name = f"{gpu_count}x {gpu_name} {cpu_cores}x {cpu_model}"

        offer = RawCatalogItem(
            instance_name=instance_name,
            location="us-michigan-1",  # Hardcoded for now, as HotAisle only has one location.
            price=price,
            cpu=cpu_cores,
            memory=memory_gb,
            gpu_count=gpu_count,
            gpu_name=gpu_name,
            gpu_memory=gpu_memory,
            gpu_vendor=gpu_vendor,
            spot=False,
            disk_size=disk_gb,
        )
        offers.append(offer)

    return offers



# ruff: noqa: F401
import warnings

from gpuhunt._internal.catalog import Catalog
from gpuhunt._internal.constraints import (
    KNOWN_ACCELERATORS,
    KNOWN_AMD_GPUS,
    KNOWN_INTEL_ACCELERATORS,
    KNOWN_NVIDIA_GPUS,
    KNOWN_TENSTORRENT_ACCELERATORS,
    KNOWN_TPUS,
    correct_gpu_memory_gib,
    is_nvidia_superchip,
    matches,
)
from gpuhunt._internal.default import default_catalog, query
from gpuhunt._internal.models import (
    AcceleratorInfo,
    AcceleratorVendor,
    AMDGPUInfo,
    CatalogItem,
    CPUArchitecture,
    IntelAcceleratorInfo,
    NvidiaGPUInfo,
    QueryFilter,
    RawCatalogItem,
    TenstorrentAcceleratorInfo,
    TPUInfo,
)

# Deprecated aliases
GPUInfo: type[NvidiaGPUInfo]
KNOWN_GPUS: list[NvidiaGPUInfo]


def _warn_renamed(old: str, new: str) -> None:
    warnings.warn(
        f"{old} has been renamed to {new}, the old name is deprecated and will be removed.",
        DeprecationWarning,
        stacklevel=2,
    )


def __getattr__(name):
    if name == "GPUInfo":
        _warn_renamed("GPUInfo", "NvidiaGPUInfo")
        return NvidiaGPUInfo
    if name == "KNOWN_GPUS":
        _warn_renamed("KNOWN_GPUS", "KNOWN_NVIDIA_GPUS")
        return KNOWN_NVIDIA_GPUS
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



import argparse
import logging
from collections.abc import Sequence
from pathlib import Path
from textwrap import dedent
from typing import Optional

from gpuhunt._internal import storage
from gpuhunt._internal.utils import configure_logging


def main(args: Optional[Sequence[str]] = None):
    configure_logging()
    parser = argparse.ArgumentParser(
        description=dedent(
            """
            Convert a v2 catalog to a v1 catalog. Legacy v1 catalogs are used by older
            gpuhunt versions that do not respect the `flags` field. Any catalog items
            with flags are filtered out when converting to v1.
            """
        )
    )
    parser.add_argument("--input", type=Path, required=True, help="The v2 catalog file to read")
    parser.add_argument("--output", type=Path, required=True, help="The v1 catalog file to write")
    args = parser.parse_args(args)
    storage.convert_catalog_v2_to_v1(path_v2=args.input, path_v1=args.output)
    logging.info("Converted %s -> %s", args.input, args.output)


if __name__ == "__main__":
    main()



import csv
import os
from dataclasses import fields
from pathlib import Path

import pytest

from gpuhunt._internal.models import RawCatalogItem

# Fields that are allowed to be empty, including empty strings or empty lists
OPTIONAL_CATALOG_ITEM_FIELDS = ["gpu_name", "gpu_memory", "gpu_vendor", "disk_size", "flags"]

files = sorted(Path(os.environ["CATALOG_DIR"]).glob("*.csv"))


def catalog_name(catalog) -> str:
    return catalog.name


class TestAllCatalogs:
    @pytest.fixture(params=files, ids=catalog_name)
    def catalog(self, request):
        yield csv.DictReader(request.param.open())

    @pytest.mark.parametrize(
        "field",
        [f.name for f in fields(RawCatalogItem) if f.name not in OPTIONAL_CATALOG_ITEM_FIELDS],
    )
    def test_field_present(self, catalog: csv.DictReader, field: str) -> None:
        for row in catalog:
            assert row[field], str(row)

    @pytest.mark.parametrize("field", ["spot"])
    def test_boolean_field(self, catalog: csv.DictReader, field: str) -> None:
        for row in catalog:
            assert row[field] in ("True", "False"), str(row)

    def test_gpu_consistent(self, catalog: csv.DictReader) -> None:
        for row in catalog:
            if int(row["gpu_count"]) > 0:
                assert row["gpu_name"] and row["gpu_memory"] and row["gpu_vendor"], str(row)
            else:
                assert not (row["gpu_name"] or row["gpu_memory"] or row["gpu_vendor"]), str(row)

    def test_price_positive(self, catalog: csv.DictReader) -> None:
        for row in catalog:
            if "gcp-g4-preview" in row["flags"]:
                continue
            assert float(row["price"]) > 0, str(row)



from typing import Union

import pytest

from gpuhunt._internal.constraints import KNOWN_AMD_GPUS
from gpuhunt._internal.models import (
    AcceleratorVendor,
    AMDArchitecture,
    CatalogItem,
    CPUArchitecture,
    Optional,
    RawCatalogItem,
)

NVIDIA = AcceleratorVendor.NVIDIA
GOOGLE = AcceleratorVendor.GOOGLE
AMD = AcceleratorVendor.AMD


@pytest.mark.parametrize(
    ["gpu_count", "gpu_vendor", "gpu_name", "expected_gpu_vendor", "expected_gpu_name"],
    [
        pytest.param(None, None, None, None, None, id="none-gpu"),
        pytest.param(0, None, None, None, None, id="zero-gpu"),
        pytest.param(1, None, "A100", "nvidia", "A100", id="one-gpu"),
        pytest.param(1, None, "tpu-v3", "google", "v3", id="one-tpu-vendor-not-set"),
        pytest.param(1, "google", "tpu-v5p", "google", "v5p", id="one-tpu-vendor-is-set"),
        pytest.param(1, AMD, "MI300X", "amd", "MI300X", id="cast-enum-to-string"),
    ],
)
def test_raw_catalog_item_gpu_vendor_heuristic(
    gpu_count: Optional[int],
    gpu_vendor: Union[AcceleratorVendor, str, None],
    gpu_name: Optional[str],
    expected_gpu_vendor: Optional[str],
    expected_gpu_name: Optional[str],
):
    dct = {}
    if gpu_vendor is not None:
        dct["gpu_vendor"] = gpu_vendor
    if gpu_count is not None:
        dct["gpu_count"] = gpu_count
    if gpu_name is not None:
        dct["gpu_name"] = gpu_name

    item = RawCatalogItem.from_dict(dct)

    assert item.gpu_vendor == expected_gpu_vendor
    assert item.gpu_name == expected_gpu_name


@pytest.mark.parametrize(
    ["gpu_count", "gpu_vendor", "gpu_name", "expected_gpu_vendor"],
    [
        pytest.param(None, None, None, None, id="none-gpu"),
        pytest.param(0, None, None, None, id="zero-gpu"),
        pytest.param(1, None, None, NVIDIA, id="one-gpu-no-name"),
        pytest.param(1, None, "v3", NVIDIA, id="one-gpu-with-any-name"),
        pytest.param(1, "amd", "MI300X", AMD, id="cast-string-to-enum"),
    ],
)
def test_catalog_item_gpu_vendor_heuristic(
    gpu_count: Optional[int],
    gpu_vendor: Union[AcceleratorVendor, str, None],
    gpu_name: Optional[str],
    expected_gpu_vendor: Optional[AcceleratorVendor],
):
    item = CatalogItem(
        instance_name="test-instance",
        location="eu-west-1",
        price=1.0,
        cpu=1,
        memory=32.0,
        gpu_vendor=gpu_vendor,
        gpu_count=gpu_count,
        gpu_name=gpu_name,
        gpu_memory=8.0,
        spot=False,
        disk_size=100.0,
        provider="test",
    )

    assert item.gpu_vendor == expected_gpu_vendor


@pytest.mark.parametrize(
    ["cpu_arch", "expected_cpu_arch"],
    [
        pytest.param(None, CPUArchitecture.X86, id="non-set"),
        pytest.param(CPUArchitecture.X86, CPUArchitecture.X86, id="enum"),
        pytest.param("ARM", CPUArchitecture.ARM, id="cast-string-to-enum"),
    ],
)
def test_catalog_item_cpu_arch_heuristic(
    cpu_arch: Union[CPUArchitecture, str, None],
    expected_cpu_arch: CPUArchitecture,
):
    item = CatalogItem(
        instance_name="test-instance",
        location="eu-west-1",
        price=1.0,
        cpu_arch=cpu_arch,
        cpu=1,
        memory=32.0,
        gpu_count=0,
        gpu_name=None,
        gpu_memory=8.0,
        spot=False,
        disk_size=100.0,
        provider="test",
    )

    assert item.cpu_arch == expected_cpu_arch


@pytest.mark.parametrize(
    ["model", "architecture", "expected_memory"],
    [
        pytest.param("MI325X", AMDArchitecture.CDNA3, 288, id="MI325X"),
        pytest.param("MI308X", AMDArchitecture.CDNA3, 128, id="MI308X"),
        pytest.param("MI300X", AMDArchitecture.CDNA3, 192, id="MI300X"),
        pytest.param("MI300A", AMDArchitecture.CDNA3, 128, id="MI300A"),
        pytest.param("MI250X", AMDArchitecture.CDNA2, 128, id="MI250X"),
        pytest.param("MI250", AMDArchitecture.CDNA2, 128, id="MI250"),
        pytest.param("MI210", AMDArchitecture.CDNA2, 64, id="MI210"),
        pytest.param("MI100", AMDArchitecture.CDNA, 32, id="MI100"),
    ],
)
def test_amd_gpu_architecture(model: str, architecture: AMDArchitecture, expected_memory: int):
    for gpu in KNOWN_AMD_GPUS:
        if gpu.name == model:
            assert gpu.architecture == architecture
            assert gpu.memory == expected_memory
            return
    # If we get here, the test should fail since we could not find the GPU in our known list.
    assert False


def test_raw_catalog_item_to_from_dict() -> None:
    item = RawCatalogItem(
        instance_name="test-instance",
        location="eu-west-1",
        price=1.0,
        cpu_arch=CPUArchitecture.ARM,
        cpu=1,
        memory=32.0,
        gpu_vendor=AcceleratorVendor.NVIDIA,
        gpu_count=1,
        gpu_name="A10",
        gpu_memory=24.0,
        spot=False,
        disk_size=100.0,
        flags=["f1", "f2", "f3"],
        provider_data={"custom_prop": 42},
    )
    item_dict = item.dict()
    assert item_dict == {
        "instance_name": "test-instance",
        "location": "eu-west-1",
        "price": 1.0,
        "cpu_arch": "arm",
        "cpu": 1,
        "memory": 32.0,
        "gpu_vendor": "nvidia",
        "gpu_count": 1,
        "gpu_name": "A10",
        "gpu_memory": 24.0,
        "spot": False,
        "disk_size": 100.0,
        "flags": "f1 f2 f3",
        "provider_data": '{"custom_prop": 42}',
    }
    assert RawCatalogItem.from_dict(item_dict) == item



import pytest

import gpuhunt._internal.catalog as internal_catalog
from gpuhunt import Catalog
from gpuhunt.providers.cudo import (
    CudoProvider,
    get_balanced_disk_size,
    get_balanced_memory,
    get_memory,
    gpu_name,
)


@pytest.fixture
def machine_types() -> list[dict]:
    return [
        {
            "dataCenterId": "br-saopaulo-1",
            "machineType": "cascade-lake",
            "cpuModel": "Cascadelake-Server-noTSX",
            "gpuModel": "RTX 3080",
            "gpuModelId": "nvidia-rtx-3080",
            "minVcpuPerMemoryGib": 0.25,
            "maxVcpuPerMemoryGib": 1,
            "minVcpuPerGpu": 1,
            "maxVcpuPerGpu": 13,
            "vcpuPriceHr": {"value": "0.002500"},
            "memoryGibPriceHr": {"value": "0.003800"},
            "gpuPriceHr": {"value": "0.05"},
            "minStorageGibPriceHr": {"value": "0.00013"},
            "ipv4PriceHr": {"value": "0.005500"},
            "maxVcpuFree": 76,
            "totalVcpuFree": 377,
            "maxMemoryGibFree": 227,
            "totalMemoryGibFree": 1132,
            "maxGpuFree": 5,
            "totalGpuFree": 24,
            "maxStorageGibFree": 42420,
            "totalStorageGibFree": 42420,
        },
        {
            "dataCenterId": "no-luster-1",
            "machineType": "epyc-rome-rtx-a5000",
            "cpuModel": "EPYC-Rome",
            "gpuModel": "RTX A5000",
            "gpuModelId": "nvidia-rtx-a5000",
            "minVcpuPerMemoryGib": 0.259109,
            "maxVcpuPerMemoryGib": 1.036437,
            "minVcpuPerGpu": 1,
            "maxVcpuPerGpu": 16,
            "vcpuPriceHr": {"value": "0.002100"},
            "memoryGibPriceHr": {"value": "0.003400"},
            "gpuPriceHr": {"value": "0.520000"},
            "minStorageGibPriceHr": {"value": "0.000107"},
            "ipv4PriceHr": {"value": "0.003500"},
            "renewableEnergy": False,
            "maxVcpuFree": 116,
            "totalVcpuFree": 208,
            "maxMemoryGibFree": 219,
            "totalMemoryGibFree": 390,
            "maxGpuFree": 4,
            "totalGpuFree": 7,
            "maxStorageGibFree": 1170,
            "totalStorageGibFree": 1170,
        },
    ]


def test_get_offers_with_query_filter(mocker, machine_types):
    catalog = Catalog(balance_resources=False, auto_reload=False)
    cudo = CudoProvider()
    cudo.list_vm_machine_types = mocker.Mock(return_value=machine_types)
    internal_catalog.ONLINE_PROVIDERS = ["cudo"]
    internal_catalog.OFFLINE_PROVIDERS = []
    catalog.add_provider(cudo)
    query_result = catalog.query(provider=["cudo"], min_gpu_count=1, max_gpu_count=1)
    assert len(query_result) >= 1, "No offers found"


def test_get_offers_for_gpu_name(mocker, machine_types):
    catalog = Catalog(balance_resources=True, auto_reload=False)
    cudo = CudoProvider()
    cudo.list_vm_machine_types = mocker.Mock(return_value=machine_types)
    internal_catalog.ONLINE_PROVIDERS = ["cudo"]
    internal_catalog.OFFLINE_PROVIDERS = []
    catalog.add_provider(cudo)
    query_result = catalog.query(provider=["cudo"], min_gpu_count=1, gpu_name=["A5000"])
    assert len(query_result) >= 1, "No offers found"


def test_get_offers_for_gpu_memory(mocker, machine_types):
    catalog = Catalog(balance_resources=True, auto_reload=False)
    cudo = CudoProvider()
    cudo.list_vm_machine_types = mocker.Mock(return_value=machine_types)
    internal_catalog.ONLINE_PROVIDERS = ["cudo"]
    internal_catalog.OFFLINE_PROVIDERS = []
    catalog.add_provider(cudo)
    query_result = catalog.query(provider=["cudo"], min_gpu_count=1, min_gpu_memory=16)
    assert len(query_result) >= 1, "No offers found"


def test_get_offers_for_compute_capability(mocker, machine_types):
    catalog = Catalog(balance_resources=True, auto_reload=False)
    cudo = CudoProvider()
    cudo.list_vm_machine_types = mocker.Mock(return_value=machine_types)
    internal_catalog.ONLINE_PROVIDERS = ["cudo"]
    internal_catalog.OFFLINE_PROVIDERS = []
    catalog.add_provider(cudo)
    query_result = catalog.query(provider=["cudo"], min_gpu_count=1, min_compute_capability=(8, 6))
    assert len(query_result) >= 1, "No offers found"


def test_get_offers_no_query_filter(mocker, machine_types):
    catalog = Catalog(balance_resources=True, auto_reload=False)
    cudo = CudoProvider()
    cudo.list_vm_machine_types = mocker.Mock(return_value=machine_types)
    internal_catalog.ONLINE_PROVIDERS = ["cudo"]
    internal_catalog.OFFLINE_PROVIDERS = []
    catalog.add_provider(cudo)
    query_result = catalog.query(provider=["cudo"])
    assert len(query_result) >= 1, "No offers found"


def test_optimize_offers_2(mocker, machine_types):
    catalog = Catalog(balance_resources=True, auto_reload=False)
    cudo = CudoProvider()
    cudo.list_vm_machine_types = mocker.Mock(return_value=machine_types[0:1])
    internal_catalog.ONLINE_PROVIDERS = ["cudo"]
    internal_catalog.OFFLINE_PROVIDERS = []
    catalog.add_provider(cudo)
    query_result = catalog.query(
        provider=["cudo"], min_cpu=2, min_gpu_count=1, max_gpu_count=1, min_memory=8
    )
    machine_type = machine_types[0]
    balance_resource = True
    available_disk = machine_type["maxStorageGibFree"]
    gpu_memory = get_memory(gpu_name(machine_type["gpuModel"]))
    max_memory = None
    max_disk_size = None
    min_disk_size = None

    assert len(query_result) >= 1

    for config in query_result:
        min_cpus_for_memory = machine_type["minVcpuPerMemoryGib"] * config.cpu
        max_cpus_for_memory = machine_type["maxVcpuPerMemoryGib"] * config.memory
        min_cpus_for_gpu = machine_type["minVcpuPerGpu"] * config.gpu_count
        assert config.cpu >= min_cpus_for_memory, (
            f"VM config does not meet the minimum CPU:Memory requirement. Required minimum CPUs: "
            f"{min_cpus_for_memory}, Found: {config.cpu}"
        )
        assert config.cpu <= max_cpus_for_memory, (
            f"VM config exceeds the maximum CPU:Memory allowance. Allowed maximum CPUs: "
            f"{max_cpus_for_memory}, Found: {config.cpu}"
        )
        assert config.cpu >= min_cpus_for_gpu, (
            f"VM config does not meet the minimum CPU:GPU requirement. "
            f"Required minimum CPUs: {min_cpus_for_gpu}, Found: {config.cpu}"
        )
        # Perform the balance resource checks if balance_resource is True
        if balance_resource:
            expected_memory = get_balanced_memory(config.gpu_count, gpu_memory, max_memory)
            expected_disk_size = get_balanced_disk_size(
                available_disk,
                config.memory,
                config.gpu_count * gpu_memory,
                max_disk_size,
                min_disk_size,
            )

            assert config.memory == expected_memory, (
                f"Memory allocation does not match the expected balanced memory. "
                f"Expected: {expected_memory}, Found: {config.memory}"
            )
            assert config.disk_size == expected_disk_size, (
                f"Disk size allocation does not match the expected balanced disk size. "
                f"Expected: {expected_disk_size}, Found: {config.disk_size}"
            )



import importlib
import inspect
import pkgutil
import sys

import pytest

import gpuhunt.providers
from gpuhunt._internal.catalog import OFFLINE_PROVIDERS, ONLINE_PROVIDERS


@pytest.fixture()
def providers():
    """List of all provider classes"""
    members = []
    for module_info in pkgutil.walk_packages(gpuhunt.providers.__path__):
        if sys.version_info < (3, 10) and module_info.name == "nebius":
            continue
        module = importlib.import_module(
            f".{module_info.name}",
            package="gpuhunt.providers",
        )
        for _, member in inspect.getmembers(module):
            if not inspect.isclass(member):
                continue
            if member.__name__.islower():
                continue  # skip builtins to avoid CPython bug #89489 in `issubclass` below
            if not issubclass(member, gpuhunt.providers.AbstractProvider):
                continue
            if member.__name__ == "AbstractProvider":
                continue
            members.append(member)
    assert members
    return members


def test_catalog_providers_is_unique():
    CATALOG_PROVIDERS = OFFLINE_PROVIDERS + ONLINE_PROVIDERS
    assert len(set(CATALOG_PROVIDERS)) == len(CATALOG_PROVIDERS)


def test_all_providers_have_a_names(providers):
    names = [p.NAME for p in providers]
    assert gpuhunt.providers.AbstractProvider.NAME not in names
    assert len(set(names)) == len(names)


def test_catalog_providers(providers):
    CATALOG_PROVIDERS = OFFLINE_PROVIDERS + ONLINE_PROVIDERS
    if sys.version_info < (3, 10):
        CATALOG_PROVIDERS = [p for p in CATALOG_PROVIDERS if p != "nebius"]
    names = [p.NAME for p in providers]
    assert set(CATALOG_PROVIDERS) == set(names)
    assert len(CATALOG_PROVIDERS) == len(names)



import pytest

from gpuhunt._internal.utils import to_camel_case


@pytest.mark.parametrize(
    ["before", "after"],
    [
        ["spam_ham_eggs", "spamHamEggs"],
        ["spam__ham__eggs", "spamHamEggs"],
        ["__spam_ham_eggs__", "spamHamEggs"],
        ["spamHam_eggs", "spamHamEggs"],
        ["spamHamEggs", "spamHamEggs"],
        ["SpamHam_eggs", "SpamHamEggs"],
        ["spam", "spam"],
        ["", ""],
    ],
)
def test_to_camel_case(before, after):
    assert to_camel_case(before) == after



import logging
import sys
from typing import Callable, Optional, Union


def configure_logging() -> None:
    logging.basicConfig(
        level=logging.INFO,
        stream=sys.stdout,
        format="%(asctime)s %(levelname)s %(message)s",
    )


def empty_as_none(value: Optional[str], loader: Optional[Callable] = None):
    if value is None or value == "":
        return None
    if loader is not None:
        return loader(value)
    return value


def parse_compute_capability(
    value: Optional[Union[str, tuple[int, int]]],
) -> Optional[tuple[int, int]]:
    if isinstance(value, str):
        major, minor = value.split(".")
        return int(major), int(minor)
    return value


def to_camel_case(snake_case: str) -> str:
    words = snake_case.split("_")
    words = list(filter(None, words))
    words[1:] = [word[:1].upper() + word[1:] for word in words[1:]]
    return "".join(words)



import csv
from collections import Counter
from pathlib import Path

import pytest

from gpuhunt.providers.datacrunch import ALL_AMD_GPUS, GPU_MAP


@pytest.fixture
def data_rows(catalog_dir: Path) -> list[dict]:
    file = catalog_dir / "datacrunch.csv"
    reader = csv.DictReader(file.open())
    return list(reader)


def select_row(rows, name: str) -> list[str]:
    return [r[name] for r in rows if r[name]]


def test_locations(data_rows):
    expected = {
        "FIN-01",
        "FIN-02",
        "FIN-02",
        "ICE-01",
    }
    locations = select_row(data_rows, "location")
    missing = expected - set(locations)
    assert not missing

    count = Counter(locations)
    for loc in expected:
        assert count[loc] > 1


def test_spot(data_rows):
    spots = select_row(data_rows, "spot")

    count = Counter(spots)
    for spot_key in ("True", "False"):
        assert count[spot_key] > 1


def test_gpu_present(data_rows):
    refs = [name for name in GPU_MAP.values() if name not in ALL_AMD_GPUS]
    gpus = select_row(data_rows, "gpu_name")
    assert set(gpus) == set(refs)



from abc import ABC, abstractmethod
from typing import Optional

from gpuhunt._internal.models import QueryFilter, RawCatalogItem


class AbstractProvider(ABC):
    """
    Abstract class for cloud provider implementations.

    Attributes:
        NAME: (class variable) The name of the provider.
    """

    NAME: str = "abstract"  # Override in subclasses

    @abstractmethod
    def get(
        self, query_filter: Optional[QueryFilter] = None, balance_resources: bool = True
    ) -> list[RawCatalogItem]:
        """
        Return a list of available instance offers. Offers should be ordered by priority. In most
        cases - by price, ascending.

        Args:
            query_filter: Set of filters requested by the user. Only used with online providers.
                Filters are safe to ignore, as they are also enforced by `gpuhunt` after calling
                `get`. However, they can be used to reduce the number or size of API requests if
                the provider's API supports filtering by GPU, RAM, region, etc.
            balance_resources: Whether the instance resources (CPU, RAM, disk) should be
                adjusted to better match the GPU. Only used with online providers. Only relevant
                to cloud providers that allow configuring instance CPU, RAM, and disk.
        """

        pass

    @classmethod
    def filter(cls, offers: list[RawCatalogItem]) -> list[RawCatalogItem]:
        """
        Return a subset of offers that should be stored in the catalog.

        Only used with offline providers. Only implement this method if there are reasons to omit
        some offers from the catalog.
        """

        return offers



import copy
import itertools
import logging
import re
from collections.abc import Iterable
from typing import Optional

from datacrunch import DataCrunchClient
from datacrunch.instance_types.instance_types import InstanceType

from gpuhunt import QueryFilter, RawCatalogItem
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)

AMD_RX7900XTX = "RX7900XTX"
ALL_AMD_GPUS = [
    AMD_RX7900XTX,
]


class DataCrunchProvider(AbstractProvider):
    NAME = "datacrunch"

    def __init__(self, client_id: str, client_secret: str) -> None:
        self.datacrunch_client = DataCrunchClient(client_id, client_secret)

    def get(
        self, query_filter: Optional[QueryFilter] = None, balance_resources: bool = True
    ) -> list[RawCatalogItem]:
        instance_types = self._get_instance_types()
        locations = self._get_locations()

        spots = (True, False)
        location_codes = [loc["code"] for loc in locations]
        instances = generate_instances(spots, location_codes, instance_types)

        return sorted(instances, key=lambda x: x.price)

    def _get_instance_types(self) -> list[InstanceType]:
        return self.datacrunch_client.instance_types.get()

    def _get_locations(self) -> list[dict]:
        return self.datacrunch_client.locations.get()

    @classmethod
    def filter(cls, offers: list[RawCatalogItem]) -> list[RawCatalogItem]:
        return [o for o in offers if o.gpu_name not in ALL_AMD_GPUS]  # skip AMD GPU


def generate_instances(
    spots: Iterable[bool], location_codes: Iterable[str], instance_types: Iterable[InstanceType]
) -> list[RawCatalogItem]:
    instances = []
    for spot, location, instance in itertools.product(spots, location_codes, instance_types):
        item = transform_instance(copy.copy(instance), spot, location)
        if item is None:
            continue
        instances.append(RawCatalogItem.from_dict(item))
    return instances


def transform_instance(instance: InstanceType, spot: bool, location: str) -> Optional[dict]:
    gpu_memory = None
    gpu_count = instance.gpu["number_of_gpus"]
    gpu_name = None

    if instance.gpu["number_of_gpus"]:
        gpu_memory = instance.gpu_memory["size_in_gigabytes"] / instance.gpu["number_of_gpus"]
        gpu_name = get_gpu_name(instance.gpu["description"])

    if gpu_count and gpu_name is None:
        logger.warning(
            "Failed to get GPU name from description: '%s'", instance.gpu["description"]
        )
        return None

    raw = dict(
        instance_name=instance.instance_type,
        location=location,
        spot=spot,
        price=instance.spot_price_per_hour if spot else instance.price_per_hour,
        cpu=instance.cpu["number_of_cores"],
        memory=instance.memory["size_in_gigabytes"],
        gpu_count=gpu_count,
        gpu_name=gpu_name,
        gpu_memory=gpu_memory,
    )
    return raw


GPU_MAP = {
    r"\d+x B200 SXM6 180GB": "B200",
    r"\d+x H200 SXM5 141GB": "H200",
    r"\d+x H100 SXM5 80GB": "H100",
    r"\d+x A100 SXM4 80GB": "A100",
    r"\d+x A100 SXM4 40GB": "A100",
    r"\d+x RTX6000 Ada 48GB": "RTX6000Ada",
    r"\d+x RTX 6000 Ada 48GB": "RTX6000Ada",
    r"\d+x RTX A6000 48GB": "A6000",
    r"\d+x Tesla V100 16GB": "V100",
    r"\d+x L40S 48GB": "L40S",
    r"\d+x AMD 7900XTX": AMD_RX7900XTX,
}


def get_gpu_name(name: str) -> Optional[str]:
    for regex, gpu_name in GPU_MAP.items():
        if re.fullmatch(regex, name):
            return gpu_name
    return None



import copy
import logging
import re
from collections import defaultdict
from typing import Any, Literal, Optional, Union

import requests

from gpuhunt._internal.constraints import correct_gpu_memory_gib
from gpuhunt._internal.models import QueryFilter, RawCatalogItem
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)
bundles_url = "https://console.vast.ai/api/v0/bundles/"
kilo = 1000
# Maximum number of offers to fetch when GPU name mapping fails.
Operators = Literal["lt", "lte", "eq", "gte", "gt"]
FilterValue = Union[int, float, str, bool]


class VastAIProvider(AbstractProvider):
    NAME = "vastai"

    def __init__(self, extra_filters: Optional[dict[str, dict[Operators, FilterValue]]] = None):
        self.extra_filters = extra_filters

    def get(
        self, query_filter: Optional[QueryFilter] = None, balance_resources: bool = True
    ) -> list[RawCatalogItem]:
        filters: dict[str, Any] = self.make_filters(query_filter or QueryFilter())
        if self.extra_filters:
            for key, constraints in self.extra_filters.items():
                for op, value in constraints.items():
                    filters[key][op] = value
        resp = requests.post(bundles_url, json=filters, timeout=10)
        resp.raise_for_status()
        data = resp.json()

        instance_offers = []
        for offer in data["offers"]:
            cpu_cores = offer["cpu_cores"]
            # although this is not stated in the docs, the value can be None
            if cpu_cores:
                memory = float(
                    int(offer["cpu_ram"] * offer["cpu_cores_effective"] / cpu_cores / kilo)
                )
            else:
                memory = 0.0
            disk_size = query_filter and query_filter.min_disk_size or offer["disk_space"]
            if not self.satisfies_filters(offer, filters):
                logger.warning("Offer %s does not satisfy filters", offer["id"])
                continue
            gpu_name = get_dstack_gpu_name(offer["gpu_name"])
            gpu_memory = correct_gpu_memory_gib(gpu_name, offer["gpu_ram"])
            ondemand_offer = RawCatalogItem(
                instance_name=str(offer["id"]),
                location=get_location(offer["geolocation"]),
                # storage_cost is $/gb/month
                price=round(
                    offer["dph_base"] + disk_size * offer["storage_cost"] / 30 / 24,
                    5,
                ),
                cpu=int(offer["cpu_cores_effective"]),
                memory=memory,
                gpu_vendor=None,
                gpu_count=offer["num_gpus"],
                gpu_name=gpu_name,
                gpu_memory=float(gpu_memory),
                spot=False,
                disk_size=disk_size,
            )
            instance_offers.append(ondemand_offer)

            if offer.get("min_bid"):
                spot_offer = copy.deepcopy(ondemand_offer)
                spot_offer.price = round(offer["min_bid"], 5)
                spot_offer.spot = True
                instance_offers.append(spot_offer)
        return instance_offers

    @staticmethod
    def make_filters(q: QueryFilter) -> dict[str, dict[Operators, FilterValue]]:
        filters = defaultdict(dict)
        if q.min_cpu is not None:
            filters["cpu_cores"]["gte"] = q.min_cpu
        if q.max_cpu is not None:
            filters["cpu_cores"]["lte"] = q.max_cpu
        if q.min_memory is not None:
            filters["cpu_ram"]["gte"] = q.min_memory * kilo
        if q.max_memory is not None:
            filters["cpu_ram"]["lte"] = q.max_memory * kilo
        if q.min_gpu_count is not None:
            filters["num_gpus"]["gte"] = q.min_gpu_count
        if q.max_gpu_count is not None:
            filters["num_gpus"]["lte"] = q.max_gpu_count
        if q.gpu_name:
            vastai_gpu_names = []
            for g in q.gpu_name:
                vastai_gpu_names.extend(get_vastai_gpu_names(g))
            if vastai_gpu_names:
                filters["gpu_name"]["in"] = vastai_gpu_names
            else:
                # If GPU name mapping fails, fetch all offers (to filter locally)
                filters["limit"] = 3000
        # See correct_gpu_memory_gib in gpuhunt/_internal/constraints.py
        if q.min_gpu_memory is not None:
            filters["gpu_ram"]["gte"] = q.min_gpu_memory * 1024 * 0.93
        if q.max_gpu_memory is not None:
            filters["gpu_ram"]["lte"] = q.max_gpu_memory * 1024 * 1.07
        if q.min_disk_size is not None:
            filters["disk_space"]["gte"] = q.min_disk_size
        if q.max_disk_size is not None:
            filters["disk_space"]["lte"] = q.max_disk_size
        if q.min_price is not None:
            filters["dph_total"]["gte"] = q.min_price
        if q.max_price is not None:
            filters["dph_total"]["lte"] = q.max_price
        # TODO(egor-s): add compute capability info for all GPUs
        if q.min_compute_capability is not None:
            filters["compute_capability"]["gte"] = compute_cap(q.min_compute_capability)
        if q.max_compute_capability is not None:
            filters["compute_capability"]["lte"] = compute_cap(q.max_compute_capability)
        filters["rentable"]["eq"] = True
        filters["rented"]["eq"] = False
        filters["order"] = [["score", "desc"]]
        return filters

    @staticmethod
    def satisfies_filters(offer: dict, filters: dict[str, dict[Operators, FilterValue]]) -> bool:
        for key in filters:
            if key not in offer:
                continue
            for op, value in filters[key].items():
                if op == "lt" and offer[key] >= value:
                    return False
                if op == "lte" and offer[key] > value:
                    return False
                if op == "eq" and offer[key] != value:
                    return False
                if op == "gte" and offer[key] < value:
                    return False
                if op == "gt" and offer[key] <= value:
                    return False
        return True


GPU_MAPPING = {
    "L40S": ["L40S"],
    "L40": ["L40"],
    "A10": ["A10"],
    "A40": ["A40"],
    "L4": ["L4"],
    "A100X": ["A100X"],
    "H200": ["H200"],
    "H200NVL": ["H200 NVL"],
    "P100": ["Tesla P100"],
    "T4": ["Tesla T4"],
    "P4": ["Tesla P4"],
    "P40": ["Tesla P40"],
    "V100": ["Tesla V100"],
    "A100": ["A100 PCIE", "A100 SXM4"],
    "A800PCIE": ["A800 PCIE"],
    "H100": ["H100 PCIE", "H100 SXM"],
    "H100NVL": ["H100 NVL"],
}

GPU_MAPPING_RULES = {
    r"^RTX(\d{4}\D?)$": r"RTX \1",  # RTX4090 -> RTX 4090, RTX4090S -> RTX 4090S
    r"^QRTX(\d{4})$": r"Q RTX \1",  # QRTX8000 -> Q RTX 8000
    r"^RTX(\d{4})Ada$": r"RTX \1Ada",  # RTX4090Ada -> RTX 4090Ada
    r"^RTX(\d{4}\D?)Ti$": r"RTX \1 Ti",  # RTX4090Ti -> RTX 4090 Ti
    r"^A(\d{4})": r"RTX A\1",  # A5000 -> RTX A5000
}


def get_vastai_gpu_names(gpu_name: str) -> list[str]:
    if gpu_name in GPU_MAPPING:
        return GPU_MAPPING[gpu_name]
    for pattern, replacement in GPU_MAPPING_RULES.items():
        if re.match(pattern, gpu_name):
            return [re.sub(pattern, replacement, gpu_name)]
    return []


def get_dstack_gpu_name(gpu_name: str) -> str:
    """
    Convert VastAI GPU names to a standardized format using essential heuristics
    """
    gpu_name = gpu_name.replace("RTX A", "A").replace("Tesla ", "")
    if gpu_name.startswith("A100 "):
        return "A100"
    if gpu_name.startswith("H100 ") and "NVL" not in gpu_name:
        return "H100"
    return gpu_name.replace(" ", "")


def get_location(location: Optional[str]) -> str:
    if location is None:
        return ""
    try:
        city, country = location.replace(", ", ",").split(",")
        location = f"{country}-{city}"
    except ValueError:
        pass
    return location.lower().replace(" ", "")


def compute_cap(cc: tuple[int, int]) -> str:
    """
    >>> compute_cap((7, 0))
    '700'
    >>> compute_cap((7, 5))
    '750'
    """
    major, minor = cc
    return f"{major}{str(minor).ljust(2, '0')}"



import functools
import importlib
import logging
from typing import Callable, TypeVar

from typing_extensions import Concatenate, ParamSpec

from gpuhunt._internal.catalog import Catalog

logger = logging.getLogger(__name__)


@functools.lru_cache
def default_catalog() -> Catalog:
    """
    Returns:
        the latest catalog with all available providers loaded
    """
    catalog = Catalog()
    catalog.load()
    for module, provider in [
        ("gpuhunt.providers.tensordock", "TensorDockProvider"),
        ("gpuhunt.providers.vastai", "VastAIProvider"),
        ("gpuhunt.providers.cudo", "CudoProvider"),
        ("gpuhunt.providers.vultr", "VultrProvider"),
        ("gpuhunt.providers.hotaisle", "HotAisleProvider"),
        ("gpuhunt.providers.digitalocean", "DigitalOceanProvider"),
    ]:
        try:
            module = importlib.import_module(module)
            provider = getattr(module, provider)()
            catalog.add_provider(provider)
        except ImportError:
            logger.warning("Failed to import provider %s", provider)
        except ValueError as e:
            # Skip providers that require missing environment variables. Eg: HotAisleProvider
            logger.warning("Skipping provider %s: %s", provider, e)
    return catalog


P = ParamSpec("P")
R = TypeVar("R")
Method = Callable[P, R]
CatalogMethod = Callable[Concatenate[Catalog, P], R]


def with_signature(method: CatalogMethod) -> Callable[[Method], Method]:
    """
    Returns:
        decorator to add the signature of the Catalog method to the decorated method
    """

    def decorator(func: Method) -> Method:
        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            return func(*args, **kwargs)

        return wrapper

    return decorator


@with_signature(Catalog.query)
def query(*args: P.args, **kwargs: P.kwargs) -> R:
    """
    Query the `default_catalog`.
    See `Catalog.query` for more details on parameters

    Returns:
        (List[CatalogItem]): the result of the query
    """
    return default_catalog().query(*args, **kwargs)



import csv
from operator import itemgetter
from pathlib import Path

import pytest

from gpuhunt.providers.cloudrift import GPU_MAP


@pytest.fixture
def data_rows(catalog_dir: Path) -> list[dict]:
    with open(catalog_dir / "cloudrift.csv") as f:
        return list(csv.DictReader(f))


def select_row(rows, name: str) -> list[str]:
    return [r[name] for r in rows if r[name]]


def test_gpu_present(data_rows: list[dict]):
    expected_gpus = [gpu for _, gpu in GPU_MAP]
    gpus = select_row(data_rows, "gpu_name")
    gpus = list(dict.fromkeys(gpus))
    assert set(gpus).issubset(
        set(expected_gpus)
    ), f"Found unexpected GPUs: {set(gpus) - set(expected_gpus)}"


# TODO: Add 3, 4, 5, ... 8
@pytest.mark.parametrize("gpu_count", [1, 2])
def test_gpu_count_present(gpu_count: int, data_rows: list[dict]):
    assert str(gpu_count) in map(itemgetter("gpu_count"), data_rows)


@pytest.mark.parametrize("location", ["us-east-nc-nr-1"])
def test_location_is_present(location: str, data_rows: list[dict]):
    assert location in map(itemgetter("location"), data_rows)



import csv
import dataclasses
from typing import TypeVar

from gpuhunt._internal.models import RawCatalogItem

CATALOG_V1_FIELDS = [
    "instance_name",
    "location",
    "price",
    "cpu",
    "memory",
    "gpu_count",
    "gpu_name",
    "gpu_memory",
    "spot",
    "disk_size",
    "gpu_vendor",
]
T = TypeVar("T", bound=RawCatalogItem)


def dump(items: list[T], path: str, *, cls: type[T] = RawCatalogItem):
    with open(path, "w", newline="") as f:
        writer = csv.DictWriter(f, fieldnames=[field.name for field in dataclasses.fields(cls)])
        writer.writeheader()
        for item in items:
            writer.writerow(item.dict())


def convert_catalog_v2_to_v1(path_v2: str, path_v1: str) -> None:
    with open(path_v2) as f_v2, open(path_v1, "w") as f_v1:
        reader = csv.DictReader(f_v2)
        writer = csv.DictWriter(f_v1, fieldnames=CATALOG_V1_FIELDS, extrasaction="ignore")
        writer.writeheader()
        for row in reader:
            if not row.get("flags"):
                writer.writerow(row)



import gpuhunt._internal.catalog as internal_catalog
from gpuhunt import Catalog
from gpuhunt.providers.vultr import VultrProvider, fetch_offers

bare_metal = {
    "plans_metal": [
        {
            "id": "vbm-256c-2048gb-8-mi300x-gpu",
            "physical_cpus": 2,
            "cpu_count": 128,
            "cpu_cores": 128,
            "cpu_threads": 256,
            "cpu_model": "EPYC 9534",
            "cpu_mhz": 2450,
            "ram": 2321924,
            "disk": 3576,
            "disk_count": 8,
            "bandwidth": 10240,
            "monthly_cost": 11773.44,
            "hourly_cost": 17.52,
            "monthly_cost_preemptible": 9891.84,
            "hourly_cost_preemptible": 14.72,
            "type": "NVMe",
            "locations": ["ord"],
        },
        {
            "id": "vbm-112c-2048gb-8-h100-gpu",
            "physical_cpus": 2,
            "cpu_count": 112,
            "cpu_cores": 112,
            "cpu_threads": 224,
            "cpu_model": "Platinum 8480+",
            "cpu_mhz": 2000,
            "ram": 2097152,
            "disk": 960,
            "disk_count": 2,
            "bandwidth": 15360,
            "monthly_cost": 16074.24,
            "hourly_cost": 23.92,
            "monthly_cost_preemptible": 12364.8,
            "hourly_cost_preemptible": 18.4,
            "type": "NVMe",
            "locations": ["sea"],
        },
    ]
}

vm_instances = {
    "plans": [
        {
            "id": "vcg-a100-1c-6g-4vram",
            "vcpu_count": 1,
            "ram": 6144,
            "disk": 70,
            "disk_count": 1,
            "bandwidth": 1024,
            "monthly_cost": 90,
            "hourly_cost": 0.123,
            "type": "vcg",
            "locations": ["ewr"],
            "gpu_vram_gb": 4,
            "gpu_type": "NVIDIA_A100",
        },
        {
            "id": "vcg-a100-12c-120g-80vram",
            "vcpu_count": 12,
            "ram": 122880,
            "disk": 1400,
            "disk_count": 1,
            "bandwidth": 10240,
            "monthly_cost": 1750,
            "hourly_cost": 2.397,
            "type": "vcg",
            "locations": ["ewr"],
            "gpu_vram_gb": 80,
            "gpu_type": "NVIDIA_A100",
        },
        {
            "id": "vcg-a100-6c-60g-40vram",
            "vcpu_count": 12,
            "ram": 61440,
            "disk": 1400,
            "disk_count": 1,
            "bandwidth": 10240,
            "monthly_cost": 800,
            "hourly_cost": 1.397,
            "type": "vcg",
            "locations": ["ewr"],
            "gpu_vram_gb": 40,
            "gpu_type": "NVIDIA_A100",
        },
    ]
}


def test_fetch_offers(requests_mock):
    # Mocking the responses for the API endpoints
    requests_mock.get("https://api.vultr.com/v2/plans-metal?per_page=500", json=bare_metal)
    requests_mock.get("https://api.vultr.com/v2/plans?type=all&per_page=500", json=vm_instances)

    # Fetch offers and verify results
    assert len(fetch_offers()) == 5
    catalog = Catalog(balance_resources=False, auto_reload=False)
    vultr = VultrProvider()
    internal_catalog.ONLINE_PROVIDERS = ["vultr"]
    internal_catalog.OFFLINE_PROVIDERS = []
    catalog.add_provider(vultr)
    assert len(catalog.query(provider=["vultr"], min_gpu_count=1, max_gpu_count=1)) == 3
    assert len(catalog.query(provider=["vultr"], min_gpu_memory=80, max_gpu_count=1)) == 1
    assert len(catalog.query(provider=["vultr"], gpu_vendor="amd")) == 1
    assert len(catalog.query(provider=["vultr"], gpu_name="MI300X")) == 1



from typing import Optional

import pytest

from gpuhunt import CatalogItem, QueryFilter
from gpuhunt._internal.constraints import correct_gpu_memory_gib, matches
from gpuhunt._internal.models import AcceleratorVendor


@pytest.fixture
def item() -> CatalogItem:
    return CatalogItem(
        instance_name="large",
        location="us-east-1",
        price=1.2,
        cpu=16,
        memory=64.0,
        gpu_vendor=AcceleratorVendor.NVIDIA,
        gpu_count=1,
        gpu_name="A100",
        gpu_memory=40.0,
        spot=False,
        provider="aws",
        disk_size=None,
    )


@pytest.fixture
def cpu_items() -> list[CatalogItem]:
    datacrunch = CatalogItem(
        instance_name="CPU.120V.480G",
        location="ICE-01",
        price=3.0,
        cpu=120,
        memory=480.0,
        gpu_vendor=None,
        gpu_count=0,
        gpu_name=None,
        gpu_memory=None,
        spot=False,
        provider="datacrunch",
        disk_size=None,
    )
    aws = CatalogItem(
        instance_name="c5.12xlarge",
        location="us-east-2",
        price=2.04,
        cpu=48,
        memory=96.0,
        gpu_vendor=None,
        gpu_count=0,
        gpu_name=None,
        gpu_memory=None,
        spot=False,
        provider="aws",
        disk_size=None,
    )
    return [datacrunch, aws]


class TestMatches:
    def test_empty(self, item: CatalogItem):
        assert matches(item, QueryFilter())

    def test_cpu(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_cpu=16))
        assert matches(item, QueryFilter(max_cpu=16))
        assert not matches(item, QueryFilter(min_cpu=32))
        assert not matches(item, QueryFilter(max_cpu=8))

    def test_memory(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_memory=64.0))
        assert matches(item, QueryFilter(max_memory=64.0))
        assert not matches(item, QueryFilter(min_memory=128.0))
        assert not matches(item, QueryFilter(max_memory=32.0))

    def test_gpu_vendor_nvidia(self, item: CatalogItem):
        assert matches(item, QueryFilter(gpu_vendor=AcceleratorVendor.NVIDIA))
        assert not matches(item, QueryFilter(gpu_vendor=AcceleratorVendor.AMD))

    def test_gpu_vendor_amd(self, item: CatalogItem):
        item.gpu_vendor = AcceleratorVendor.AMD
        assert matches(item, QueryFilter(gpu_vendor=AcceleratorVendor.AMD))
        assert not matches(item, QueryFilter(gpu_vendor=AcceleratorVendor.NVIDIA))

    def test_gpu_count(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_gpu_count=1))
        assert matches(item, QueryFilter(max_gpu_count=1))
        assert not matches(item, QueryFilter(min_gpu_count=2))
        assert not matches(item, QueryFilter(max_gpu_count=0))

    def test_gpu_memory(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_gpu_memory=40.0))
        assert matches(item, QueryFilter(max_gpu_memory=40.0))
        assert not matches(item, QueryFilter(min_gpu_memory=80.0))
        assert not matches(item, QueryFilter(max_gpu_memory=20.0))

    def test_gpu_name(self, item: CatalogItem):
        assert matches(item, QueryFilter(gpu_name=["a100"]))
        assert matches(item, QueryFilter(gpu_name=["A100"]))
        assert not matches(item, QueryFilter(gpu_name=["A10"]))

    def test_gpu_name_with_filter_setattr(self, item: CatalogItem):
        q = QueryFilter()
        q.gpu_name = ["a100"]
        assert matches(item, q)
        q.gpu_name = ["A100"]
        assert matches(item, q)
        q.gpu_name = ["A10"]
        assert not matches(item, q)

    def test_total_gpu_memory(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_total_gpu_memory=40.0))
        assert matches(item, QueryFilter(max_total_gpu_memory=40.0))
        assert not matches(item, QueryFilter(min_total_gpu_memory=80.0))
        assert not matches(item, QueryFilter(max_total_gpu_memory=20.0))

    def test_price(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_price=1.2))
        assert matches(item, QueryFilter(max_price=1.2))
        assert not matches(item, QueryFilter(min_price=1.3))
        assert not matches(item, QueryFilter(max_price=1.1))

    def test_spot(self, item: CatalogItem):
        assert matches(item, QueryFilter(spot=False))
        assert not matches(item, QueryFilter(spot=True))

    def test_compute_capability(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_compute_capability=(8, 0)))
        assert matches(item, QueryFilter(max_compute_capability=(8, 0)))
        assert not matches(item, QueryFilter(min_compute_capability=(8, 1)))
        assert not matches(item, QueryFilter(max_compute_capability=(7, 9)))

    def test_compute_capability_not_nvidia(self, item: CatalogItem):
        item.gpu_vendor = AcceleratorVendor.AMD
        assert not matches(item, QueryFilter(min_compute_capability=(8, 0)))
        assert not matches(item, QueryFilter(max_compute_capability=(8, 0)))

    def test_ti_gpu(self):
        item = CatalogItem(
            instance_name="large",
            location="us-east-1",
            price=1.2,
            cpu=16,
            memory=64.0,
            gpu_count=1,
            gpu_vendor=AcceleratorVendor.NVIDIA,
            gpu_name="RTX3060Ti",  # case-sensitive
            gpu_memory=8.0,
            spot=False,
            provider="aws",
            disk_size=None,
        )
        assert matches(item, QueryFilter(gpu_name=["RTX3060TI"]))

    def test_provider(self, cpu_items):
        assert matches(cpu_items[0], QueryFilter(provider=["datacrunch"]))
        assert matches(cpu_items[0], QueryFilter(provider=["DataCrunch"]))
        assert not matches(cpu_items[0], QueryFilter(provider=["aws"]))

        assert matches(cpu_items[1], QueryFilter(provider=["aws"]))
        assert matches(cpu_items[1], QueryFilter(provider=["AWS"]))
        assert not matches(cpu_items[1], QueryFilter(provider=["datacrunch"]))

    def test_provider_with_filter_setattr(self, cpu_items):
        q = QueryFilter()
        q.provider = ["datacrunch"]
        assert matches(cpu_items[0], q)
        q.provider = ["DataCrunch"]
        assert matches(cpu_items[0], q)
        q.provider = ["aws"]
        assert not matches(cpu_items[0], q)

    @pytest.mark.parametrize(
        ("item_flags", "query_allowed_flags", "should_match"),
        [
            pytest.param([], ["a"], True, id="matches-if-no-flags"),
            pytest.param([], None, True, id="matches-if-no-flags-and-all-flags-allowed"),
            pytest.param(["a", "b"], None, True, id="matches-if-has-flags-and-all-flags-allowed"),
            pytest.param(
                ["a", "b"], ["a", "b", "c"], True, id="matches-if-has-flags-all-of-which-allowed"
            ),
            pytest.param(["a", "b"], ["a"], False, id="not-matches-if-some-flags-not-allowed"),
        ],
    )
    def test_flags(
        self,
        item: CatalogItem,
        item_flags: list[str],
        query_allowed_flags: Optional[list[str]],
        should_match: bool,
    ) -> None:
        item.flags = item_flags
        q = QueryFilter(allowed_flags=query_allowed_flags)
        assert matches(item, q) == should_match

    def test_matches_cpu_instances_with_zero_gpu_count_and_gpu_memory(self):
        cpu_item = CatalogItem(
            instance_name="large",
            location="us-east-1",
            price=1.2,
            cpu=16,
            memory=64.0,
            gpu_vendor=None,
            gpu_count=0,
            gpu_name=None,
            gpu_memory=None,
            spot=False,
            provider="aws",
            disk_size=None,
        )
        assert matches(cpu_item, QueryFilter(min_gpu_count=0, min_gpu_memory=24))


@pytest.mark.parametrize(
    ("gpu_name", "memory_mib", "expected_memory_gib"),
    [
        ("H100NVL", 95830.0, 94),
        ("L40S", 46068.0, 48),
        ("A10G", 23028.0, 24),
        ("A10", 4096.0, 4),
        ("unknown", 8200.1, 8),
    ],
)
def test_correct_gpu_memory(gpu_name: str, memory_mib: float, expected_memory_gib: int) -> None:
    assert correct_gpu_memory_gib(gpu_name, memory_mib) == expected_memory_gib



import gpuhunt._internal.catalog as internal_catalog
from gpuhunt import Catalog
from gpuhunt.providers.digitalocean import DigitalOceanProvider

sizes_response = {
    "sizes": [
        {
            "slug": "s-1vcpu-512mb-10gb",
            "memory": 512,
            "vcpus": 1,
            "disk": 10,
            "transfer": 0.5,
            "price_monthly": 4,
            "price_hourly": 0.00595,
            "regions": ["ams3", "fra1", "nyc1", "nyc2", "sfo2", "sfo3", "sgp1", "syd1"],
            "available": True,
            "description": "Basic",
            "networking_throughput": 2000,
            "disk_info": [{"type": "local", "size": {"amount": 10, "unit": "gib"}}],
        },
        {
            "slug": "gpu-h100x8-640gb",
            "memory": 1966080,
            "vcpus": 160,
            "disk": 2046,
            "transfer": 60.0,
            "price_monthly": 17796.48,
            "price_hourly": 23.92,
            "regions": ["nyc1"],
            "available": True,
            "description": "H100 GPU - 8X",
            "networking_throughput": 10000,
            "gpu_info": {
                "count": 8,
                "vram": {"amount": 640, "unit": "gib"},
                "model": "nvidia_h100",
            },
            "disk_info": [
                {"type": "local", "size": {"amount": 2046, "unit": "gib"}},
                {"type": "scratch", "size": {"amount": 40960, "unit": "gib"}},
            ],
        },
        {
            "slug": "gpu-mi300x8-1536gb",
            "memory": 1966080,
            "vcpus": 160,
            "disk": 2046,
            "transfer": 60.0,
            "price_monthly": 11844.48,
            "price_hourly": 15.92,
            "regions": ["nyc1"],
            "available": True,
            "description": "AMD MI300X - 8X",
            "networking_throughput": 10000,
            "gpu_info": {
                "count": 8,
                "vram": {"amount": 1536, "unit": "gib"},
                "model": "amd_mi300x",
            },
            "disk_info": [
                {"type": "local", "size": {"amount": 2046, "unit": "gib"}},
                {"type": "scratch", "size": {"amount": 40960, "unit": "gib"}},
            ],
        },
    ],
    "links": {},
    "meta": {"total": 147},
}


def test_fetch_offers(requests_mock):
    requests_mock.get("https://api.digitalocean.com/v2/sizes", json=sizes_response)

    provider = DigitalOceanProvider(api_key="test-token", api_url="https://api.digitalocean.com")
    offers = provider.fetch_offers()
    assert len(offers) == 10  # 8 CPU offers (8 regions) + 1 NVIDIA + 1 AMD (1 region each)
    catalog = Catalog(balance_resources=False, auto_reload=False)
    digitalocean = DigitalOceanProvider(
        api_key="test-token", api_url="https://api.digitalocean.com"
    )
    internal_catalog.ONLINE_PROVIDERS = ["digitalocean"]
    internal_catalog.OFFLINE_PROVIDERS = []
    catalog.add_provider(digitalocean)

    # Test queries
    assert (
        len(catalog.query(provider=["digitalocean"], max_gpu_count=0)) == 8
    )  # CPU only (8 regions)
    assert len(catalog.query(provider=["digitalocean"], min_gpu_count=1)) == 2  # GPU instances
    assert len(catalog.query(provider=["digitalocean"], gpu_vendor="nvidia")) == 1  # NVIDIA GPU
    assert len(catalog.query(provider=["digitalocean"], gpu_name="H100")) == 1  # Specific GPU
    assert len(catalog.query(provider=["digitalocean"], gpu_vendor="amd")) == 1  # AMD GPU
    assert len(catalog.query(provider=["digitalocean"], gpu_name="MI300X")) == 1  # AMD GPU
    assert (
        len(catalog.query(provider=["digitalocean"], min_gpu_memory=1000)) == 1
    )  # MI300X: 1536GB



from typing import Union
from unittest.mock import Mock

import gpuhunt._internal.catalog as internal_catalog
from gpuhunt import Catalog, CatalogItem, RawCatalogItem
from gpuhunt.providers.tensordock import TensorDockProvider
from gpuhunt.providers.vastai import VastAIProvider


class TestQuery:
    def test_query_merge(self):
        catalog = Catalog(balance_resources=False, auto_reload=False)

        tensordock = TensorDockProvider()
        tensordock.get = Mock(return_value=[catalog_item(price=1), catalog_item(price=3)])
        catalog.add_provider(tensordock)

        vastai = VastAIProvider()
        vastai.get = Mock(return_value=[catalog_item(price=2), catalog_item(price=1)])
        catalog.add_provider(vastai)

        assert catalog.query(provider=["tensordock", "vastai"]) == [
            catalog_item(provider="tensordock", price=1),
            catalog_item(provider="vastai", price=2),
            catalog_item(provider="vastai", price=1),
            catalog_item(provider="tensordock", price=3),
        ]

    def test_no_providers_some_not_loaded(self):
        catalog = Catalog(balance_resources=False, auto_reload=False)

        tensordock = TensorDockProvider()
        tensordock.get = Mock(return_value=[catalog_item(price=1)])
        catalog.add_provider(tensordock)

        internal_catalog.OFFLINE_PROVIDERS = []
        assert catalog.query() == [
            catalog_item(provider="tensordock", price=1),
        ]

    def test_provider_filter(self):
        catalog = Catalog(balance_resources=False, auto_reload=False)
        catalog.add_provider(tensordock := TensorDockProvider())
        catalog.add_provider(vastai := VastAIProvider())

        tensordock_offers = [catalog_item(price=1)]
        vastai_offers = [catalog_item(price=2), catalog_item(price=3)]

        tensordock.get = Mock(return_value=tensordock_offers)
        vastai.get = Mock(return_value=vastai_offers)

        assert len(catalog.query(provider="tensordock")) == 1
        assert len(catalog.query(provider="Tensordock")) == 1
        assert len(catalog.query(provider="vastai")) == 2
        assert len(catalog.query(provider="VastAI")) == 2
        assert len(catalog.query(provider=["tensordock", "VastAI"])) == 3

    def test_gpu_name_filter(self):
        catalog = Catalog(balance_resources=False, auto_reload=False)
        catalog.add_provider(tensordock := TensorDockProvider())

        tensordock.get = Mock(
            return_value=[
                catalog_item(gpu_name="A10"),
                catalog_item(gpu_name="A100"),
                catalog_item(gpu_name="a100"),
            ]
        )

        assert len(catalog.query(gpu_name="V100")) == 0
        assert len(catalog.query(gpu_name="A10")) == 1
        assert len(catalog.query(gpu_name="a10")) == 1
        assert len(catalog.query(gpu_name="A100")) == 2
        assert len(catalog.query(gpu_name="a100")) == 2
        assert len(catalog.query(gpu_name=["a10", "A100"])) == 3


def catalog_item(**kwargs) -> Union[CatalogItem, RawCatalogItem]:
    values = dict(
        instance_name="instance",
        cpu=1,
        memory=1,
        gpu_vendor="nvidia",
        gpu_count=1,
        gpu_name="gpu",
        gpu_memory=1,
        location="location",
        price=1,
        spot=False,
        disk_size=None,
    )
    values.update(kwargs)
    if "provider" in values:
        return CatalogItem(**values)
    return RawCatalogItem(**values)



import copy
import logging
import re
from typing import Optional

from requests import Session

from gpuhunt._internal.constraints import is_nvidia_superchip
from gpuhunt._internal.models import (
    CPUArchitecture,
    QueryFilter,
    RawCatalogItem,
)
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)
INSTANCE_TYPES_URL = "https://cloud.lambdalabs.com/api/v1/instance-types"
IMAGES_URL = "https://cloud.lambdalabs.com/api/v1/images"
TIMEOUT = 10

FLAG_ARM = "lambda-arm"


class LambdaLabsProvider(AbstractProvider):
    NAME = "lambdalabs"

    def __init__(self, token: str):
        self.session = Session()
        self.session.headers.update({"Authorization": f"Bearer {token}"})

    def get(
        self, query_filter: Optional[QueryFilter] = None, balance_resources: bool = True
    ) -> list[RawCatalogItem]:
        offers = []
        resp = self.session.get(INSTANCE_TYPES_URL, timeout=TIMEOUT)
        resp.raise_for_status()
        data = resp.json()["data"]
        for instance in data.values():
            instance = instance["instance_type"]
            logger.info(instance["name"])
            description = instance["description"]
            result = parse_description(description)
            if result is None:
                logger.warning("Can't parse GPU info from description: %s", description)
                continue
            gpu_count, gpu_name, gpu_memory = result
            flags: list[str] = []
            cpu_arch = CPUArchitecture.X86
            if is_nvidia_superchip(gpu_name):
                cpu_arch = CPUArchitecture.ARM
                flags.append(FLAG_ARM)
            offer = RawCatalogItem(
                instance_name=instance["name"],
                price=instance["price_cents_per_hour"] / 100,
                cpu_arch=cpu_arch.value,
                cpu=instance["specs"]["vcpus"],
                memory=float(instance["specs"]["memory_gib"]) * 1.074,
                gpu_vendor=None,
                gpu_count=gpu_count,
                gpu_name=gpu_name,
                gpu_memory=gpu_memory,
                spot=False,
                location=None,
                disk_size=float(instance["specs"]["storage_gib"]) * 1.074,
                flags=flags,
            )
            offers.append(offer)
        offers = self.add_regions(offers)
        return sorted(offers, key=lambda i: i.price)

    def add_regions(self, offers: list[RawCatalogItem]) -> list[RawCatalogItem]:
        # TODO: we don't know which regions are actually available for each instance type
        region_offers = []
        for region in self.list_regions():
            for offer in offers:
                offer = copy.deepcopy(offer)
                offer.location = region
                region_offers.append(offer)
        return region_offers

    def list_regions(self) -> list[str]:
        resp = self.session.get(IMAGES_URL, timeout=TIMEOUT)
        resp.raise_for_status()
        regions = set()
        for image in resp.json()["data"]:
            regions.add(image["region"]["name"])
        return sorted(regions)


def parse_description(v: str) -> Optional[tuple[int, str, float]]:
    """Returns gpus count, gpu name, and GPU memory"""
    r = re.match(r"^(\d)x (?:Tesla )?(.+) \((\d+) GB", v)
    if r is None:
        return None
    count, gpu_name, gpu_memory = r.groups()
    return int(count), gpu_name.replace(" ", ""), float(gpu_memory)



import pytest

from gpuhunt.providers.oci import get_gpu_name


@pytest.mark.parametrize(
    ("shape_name", "gpu_name"),
    [
        ("VM.GPU.A10.2", "A10"),
        ("BM.GPU.A100-v2.8", "A100"),
        ("BM.GPU4.8", "A100"),
        ("VM.GPU3.4", "V100"),
        ("VM.GPU2.1", "P100"),
        ("BM.GPU.H100.8", "H100"),
        ("VM.Standard2.8", None),
        ("VM.Notgpu.A10", None),
    ],
)
def test_get_gpu_name(shape_name, gpu_name):
    assert get_gpu_name(shape_name) == gpu_name



import logging
import os
from typing import Optional, Union

import requests

from gpuhunt import QueryFilter, RawCatalogItem
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)

CLOUDRIFT_SERVER_ADDRESS = "https://api.cloudrift.ai"
CLOUDRIFT_API_VERSION = "2025-03-21"


class CloudRiftProvider(AbstractProvider):
    NAME = "cloudrift"

    def get(
        self, query_filter: Optional[QueryFilter] = None, balance_resources: bool = True
    ) -> list[RawCatalogItem]:
        instance_types = self._get_instance_types()
        instance_types = [
            inst for instance in instance_types for inst in generate_instances(instance)
        ]
        return sorted(instance_types, key=lambda x: x.price)

    def _get_instance_types(self):
        request_data = {"selector": {"ByServiceAndLocation": {"services": ["vm"]}}}
        response_data = _make_request("instance-types/list", request_data)
        return response_data["instance_types"]


def generate_instances(instance) -> list[RawCatalogItem]:
    instance_gpu_brand = instance["brand_short"]
    dstack_gpu_name = next(
        iter(gpu_record[1] for gpu_record in GPU_MAP if gpu_record[0] in instance_gpu_brand), None
    )
    if dstack_gpu_name is None:
        logger.warning(f"Failed to find GPU name matching '{instance_gpu_brand}'")
        return []

    instance_types = []
    for variant in instance["variants"]:
        for location, _count in variant["nodes_per_dc"].items():
            raw = RawCatalogItem(
                instance_name=variant["name"],
                location=location,
                spot=False,
                price=variant["cost_per_hour"] / 100,
                cpu=variant["cpu_count"],
                memory=variant["dram"] / 1024**3,
                disk_size=variant["disk"] / 1024**3,
                gpu_count=variant["gpu_count"],
                gpu_name=dstack_gpu_name,
                gpu_memory=round(variant["vram"] / 1024**3),
            )
            instance_types.append(raw)

    return instance_types


GPU_MAP = [
    ("RTX 4090", "RTX4090"),
    ("RTX 5090", "RTX5090"),
    ("RTX PRO 6000", "RTXPRO6000"),
]


def _make_request(endpoint: str, request_data: dict) -> Union[dict, str, None]:
    server = os.environ.get("CLOUDRIFT_SERVER_ADDRESS", CLOUDRIFT_SERVER_ADDRESS)
    response = requests.request(
        "POST",
        f"{server}/api/v1/{endpoint}",
        json={"version": CLOUDRIFT_API_VERSION, "data": request_data},
        timeout=5.0,
    )
    if not response.ok:
        response.raise_for_status()
    try:
        response_json = response.json()
        if isinstance(response_json, str):
            return response_json
        return response_json["data"]
    except requests.exceptions.JSONDecodeError:
        return None



import csv
from pathlib import Path

import pytest


@pytest.fixture
def data_rows(catalog_dir: Path) -> list[dict]:
    with open(catalog_dir / "azure.csv") as f:
        return list(csv.DictReader(f))


class TestAzureCatalog:
    def test_standard_d2s_v3_locations(self, data_rows: list[dict]):
        expected_locations = {
            "attatlanta1",
            "attdallas1",
            "attdetroit1",
            "attnewyork1",
            "australiacentral",
            "australiacentral2",
            "australiaeast",
            "australiasoutheast",
            "australiasoutheast",
            "brazilsouth",
            "brazilsoutheast",
            "canadacentral",
            "canadaeast",
            "centralindia",
            "centralus",
            "eastasia",
            "eastus",
            "eastus2",
            "francecentral",
            "francesouth",
            "germanynorth",
            "germanywestcentral",
            "indonesiacentral",
            "israelcentral",
            "italynorth",
            "japaneast",
            "japanwest",
            "jioindiacentral",
            "jioindiawest",
            "koreacentral",
            "koreasouth",
            "malaysiawest",
            "mexicocentral",
            "newzealandnorth",
            "northcentralus",
            "northeurope",
            "norwayeast",
            "norwaywest",
            "polandcentral",
            "qatarcentral",
            "sgxsingapore1",
            "southafricanorth",
            "southafricawest",
            "southcentralus",
            "southeastasia",
            "southindia",
            "spaincentral",
            "swedencentral",
            "swedensouth",
            "switzerlandnorth",
            "switzerlandwest",
            "uaecentral",
            "uaenorth",
            "uksouth",
            "ukwest",
            "usgovarizona",
            "usgovtexas",
            "usgovvirginia",
            "westcentralus",
            "westeurope",
            "westindia",
            "westus",
            "westus2",
            "westus3",
        }
        locations = set(
            row["location"] for row in data_rows if row["instance_name"] == "Standard_D2s_v3"
        )
        missing = expected_locations - locations
        assert not missing

    def test_spots_presented(self, data_rows: list[dict]):
        assert any(row["spot"] == "True" for row in data_rows)

    def test_ondemand_presented(self, data_rows: list[dict]):
        assert any(row["spot"] == "False" for row in data_rows)

    def test_gpu_presented(self, data_rows: list[dict]):
        expected_gpus = {
            "A100",
            "A10",
            "T4",
            "V100",
        }
        gpus = set(row["gpu_name"] for row in data_rows if row["gpu_name"])
        assert expected_gpus == gpus

    def test_both_a100_presented(self, data_rows: list[dict]):
        expected_gpu_memory = {"40.0", "80.0"}
        gpu_memory = set(row["gpu_memory"] for row in data_rows if row["gpu_name"] == "A100")
        assert expected_gpu_memory == gpu_memory



import csv
import json
from operator import itemgetter
from pathlib import Path

import pytest


@pytest.fixture
def data_rows(catalog_dir: Path) -> list[dict]:
    with open(catalog_dir / "nebius.csv") as f:
        return list(csv.DictReader(f))


@pytest.mark.parametrize("gpu", ["L40S", "H100", "H200", "B200", ""])
def test_gpu_present(gpu: str, data_rows: list[dict]):
    assert gpu in map(itemgetter("gpu_name"), data_rows)


def test_on_demand_present(data_rows: list[dict]):
    assert "False" in map(itemgetter("spot"), data_rows)


def test_spots_presented(data_rows: list[dict]):
    spot_rows = [row for row in data_rows if row["spot"] == "True"]
    assert len(spot_rows) > 0


@pytest.mark.parametrize("location", ["eu-north1", "eu-west1"])
def test_location_present(location: str, data_rows: list[dict]):
    assert location in map(itemgetter("location"), data_rows)


def test_fabrics_unique(data_rows: list[dict]) -> None:
    for row in data_rows:
        fabrics = json.loads(row["provider_data"])["fabrics"]
        assert len(fabrics) == len(set(fabrics)), f"Duplicate fabrics in row: {row}"


def test_fabrics_on_sample_offer(data_rows: list[dict]) -> None:
    for row in data_rows:
        if (
            row["instance_name"] == "gpu-h100-sxm 8gpu-128vcpu-1600gb"
            and row["location"] == "eu-north1"
        ):
            break
    else:
        raise ValueError("Offer not found")
    fabrics = set(json.loads(row["provider_data"])["fabrics"])
    expected_fabrics = {
        "fabric-2",
        "fabric-3",
        "fabric-4",
        "fabric-6",
    }
    missing_fabrics = expected_fabrics - fabrics
    assert not missing_fabrics


def test_no_fabrics_on_sample_non_clustered_offer(data_rows: list[dict]) -> None:
    for row in data_rows:
        if (
            row["instance_name"] == "gpu-h100-sxm 1gpu-16vcpu-200gb"
            and row["location"] == "eu-north1"
        ):
            break
    else:
        raise ValueError("Offer not found")
    assert json.loads(row["provider_data"])["fabrics"] == []



import re
from collections.abc import Container, Iterable
from typing import Optional, TypeVar, Union

from gpuhunt._internal.models import (
    AcceleratorInfo,
    AcceleratorVendor,
    AMDArchitecture,
    AMDGPUInfo,
    CatalogItem,
    IntelAcceleratorInfo,
    NvidiaGPUInfo,
    QueryFilter,
    TenstorrentAcceleratorInfo,
    TPUInfo,
)

# v5litepod = v5e
_TPU_VERSIONS = ["v2", "v3", "v4", "v5p", "v5litepod", "v6e"]


Comparable = TypeVar("Comparable", bound=Union[int, float, tuple[int, int]])


def is_between(value: Comparable, left: Optional[Comparable], right: Optional[Comparable]) -> bool:
    if is_below(value, left) or is_above(value, right):
        return False
    return True


def is_below(value: Comparable, limit: Optional[Comparable]) -> bool:
    if limit is not None and value < limit:
        return True
    return False


def is_above(value: Comparable, limit: Optional[Comparable]) -> bool:
    if limit is not None and value > limit:
        return True
    return False


def matches(i: CatalogItem, q: QueryFilter) -> bool:
    """
    Check if the catalog item matches the filters

    Args:
        i: catalog item
        q: filters

    Returns:
        whether the catalog item matches the filters
    """
    if q.provider is not None and i.provider.lower() not in map(str.lower, q.provider):
        return False
    if not is_between(i.price, q.min_price, q.max_price):
        return False
    if q.spot is not None and i.spot != q.spot:
        return False
    if q.cpu_arch and q.cpu_arch != i.cpu_arch:
        return False
    if not is_between(i.cpu, q.min_cpu, q.max_cpu):
        return False
    if not is_between(i.memory, q.min_memory, q.max_memory):
        return False
    if not (q.min_gpu_count == 0 and i.gpu_count == 0):
        # GPU filters should not be applied to non-gpu offers if `q.min_gpu_count == 0`.
        if q.gpu_vendor and q.gpu_vendor != i.gpu_vendor:
            return False
        if not is_between(i.gpu_count, q.min_gpu_count, q.max_gpu_count):
            return False
        if q.gpu_name is not None:
            if i.gpu_name is None:
                return False
            if i.gpu_name.lower() not in map(str.lower, q.gpu_name):
                return False
        if q.min_compute_capability is not None or q.max_compute_capability is not None:
            if i.gpu_vendor != AcceleratorVendor.NVIDIA:
                return False
            if not i.gpu_name:
                return False
            cc = get_compute_capability(i.gpu_name)
            if not cc or not is_between(cc, q.min_compute_capability, q.max_compute_capability):
                return False
        if not is_between(
            i.gpu_memory if i.gpu_count > 0 else 0, q.min_gpu_memory, q.max_gpu_memory
        ):
            return False
        if not is_between(
            (i.gpu_count * i.gpu_memory) if i.gpu_count > 0 else 0,
            q.min_total_gpu_memory,
            q.max_total_gpu_memory,
        ):
            return False
    if i.disk_size is not None:
        if not is_between(i.disk_size, q.min_disk_size, q.max_disk_size):
            return False
    if q.allowed_flags is not None:
        if any(flag not in q.allowed_flags for flag in i.flags):
            return False
    return True


def find_accelerators(
    names: Optional[Iterable[str]] = None, vendors: Optional[Container[AcceleratorVendor]] = None
) -> list[AcceleratorInfo]:
    if names is not None:
        names = {n.lower() for n in names}
    result = []
    for accelerator in KNOWN_ACCELERATORS:
        if (names is None or accelerator.name.lower() in names) and (
            vendors is None or accelerator.vendor in vendors
        ):
            result.append(accelerator)
    return result


def get_compute_capability(gpu_name: str) -> Optional[tuple[int, int]]:
    if accelerators := find_accelerators(names=[gpu_name], vendors=AcceleratorVendor.NVIDIA):
        assert isinstance(accelerators[0], NvidiaGPUInfo)
        return accelerators[0].compute_capability
    return None


def get_gpu_vendor(gpu_name: Optional[str]) -> Optional[AcceleratorVendor]:
    if gpu_name is None:
        return None
    if accelerators := find_accelerators(names=[gpu_name]):
        return accelerators[0].vendor
    return None


def correct_gpu_memory_gib(gpu_name: str, memory_mib: float) -> int:
    """
    Round to whole number of gibibytes and attempt correcting the reported GPU
    memory size if the actual memory size for that GPU is known and the
    difference between the reported and the known memory is within a heuristic
    threshold.

    This is useful for cases when nvidia-smi or cloud providers report the GPU
    memory imprecisely.
    """

    memory_gib = memory_mib / 1024
    known_memories_gib = {gpu.memory for gpu in find_accelerators(names=[gpu_name])}
    if known_memories_gib:
        closest_known_memory_gib = min(known_memories_gib, key=lambda x: abs(x - memory_gib))
        difference_gib = abs(closest_known_memory_gib - memory_gib)
        if difference_gib / closest_known_memory_gib < 0.07:
            return closest_known_memory_gib
    return round(memory_gib)


def is_nvidia_superchip(gpu_name: str) -> bool:
    """
    Check if the given NVIDIA GPU is actually a so-called "superchip" combining GPU with ARM CPU,
    such as:
    * GH200 (Grace + Hopper)
    * GB10, GB200 (Grace + Blackwell)
    """
    return re.match(r"^g[bh]\d+", gpu_name.lower()) is not None


KNOWN_NVIDIA_GPUS: list[NvidiaGPUInfo] = [
    NvidiaGPUInfo(name="A10", memory=24, compute_capability=(8, 6)),
    NvidiaGPUInfo(name="A16", memory=16, compute_capability=(8, 6)),
    NvidiaGPUInfo(name="A40", memory=48, compute_capability=(8, 6)),
    NvidiaGPUInfo(name="A100", memory=40, compute_capability=(8, 0)),
    NvidiaGPUInfo(name="A100", memory=80, compute_capability=(8, 0)),
    NvidiaGPUInfo(name="A10G", memory=24, compute_capability=(8, 6)),
    NvidiaGPUInfo(name="A4000", memory=16, compute_capability=(8, 6)),
    NvidiaGPUInfo(name="A4500", memory=20, compute_capability=(8, 6)),
    NvidiaGPUInfo(name="A5000", memory=24, compute_capability=(8, 6)),
    NvidiaGPUInfo(name="A6000", memory=48, compute_capability=(8, 6)),
    NvidiaGPUInfo(name="H100", memory=80, compute_capability=(9, 0)),
    NvidiaGPUInfo(name="H100NVL", memory=94, compute_capability=(9, 0)),
    NvidiaGPUInfo(name="H200", memory=141, compute_capability=(9, 0)),
    NvidiaGPUInfo(name="B200", memory=180, compute_capability=(10, 0)),
    NvidiaGPUInfo(name="L4", memory=24, compute_capability=(8, 9)),
    NvidiaGPUInfo(name="L40", memory=48, compute_capability=(8, 9)),
    NvidiaGPUInfo(name="L40S", memory=48, compute_capability=(8, 9)),
    NvidiaGPUInfo(name="P100", memory=16, compute_capability=(6, 0)),
    NvidiaGPUInfo(name="RTX3060", memory=8, compute_capability=(8, 6)),
    NvidiaGPUInfo(name="RTX3060", memory=12, compute_capability=(8, 6)),
    NvidiaGPUInfo(name="RTX3060Ti", memory=8, compute_capability=(8, 6)),
    NvidiaGPUInfo(name="RTX3070Ti", memory=8, compute_capability=(8, 6)),
    NvidiaGPUInfo(name="RTX3080", memory=10, compute_capability=(8, 6)),
    NvidiaGPUInfo(name="RTX3080Ti", memory=12, compute_capability=(8, 6)),
    NvidiaGPUInfo(name="RTX3090", memory=24, compute_capability=(8, 6)),
    NvidiaGPUInfo(name="RTX4090", memory=24, compute_capability=(8, 9)),
    NvidiaGPUInfo(name="RTX6000", memory=24, compute_capability=(7, 5)),
    NvidiaGPUInfo(name="RTX2000Ada", memory=16, compute_capability=(8, 9)),
    NvidiaGPUInfo(name="RTX4000Ada", memory=20, compute_capability=(8, 9)),
    NvidiaGPUInfo(name="RTX6000Ada", memory=48, compute_capability=(8, 9)),
    NvidiaGPUInfo(name="T4", memory=16, compute_capability=(7, 5)),
    NvidiaGPUInfo(name="V100", memory=16, compute_capability=(7, 0)),
    NvidiaGPUInfo(name="V100", memory=32, compute_capability=(7, 0)),
    NvidiaGPUInfo(name="GH200", memory=96, compute_capability=(9, 0)),
]

# For device ids check https://gitlab.freedesktop.org/mesa/libdrm/-/blob/main/data/amdgpu.ids
KNOWN_AMD_GPUS: list[AMDGPUInfo] = [
    AMDGPUInfo(
        name="MI100",
        memory=32,
        architecture=AMDArchitecture.CDNA,
        device_ids=(0x738C,),
    ),
    AMDGPUInfo(
        name="MI210",
        memory=64,
        architecture=AMDArchitecture.CDNA2,
        device_ids=(0x740F,),
    ),
    # TODO: recheck MI250/MI250X device ids on real devices, as the source is ambiguous:
    # 7408, 00, AMD Instinct MI250X
    # 740C, 01, AMD Instinct MI250X / MI250
    AMDGPUInfo(
        name="MI250",
        memory=128,
        architecture=AMDArchitecture.CDNA2,
        device_ids=(0x740C,),
    ),
    AMDGPUInfo(
        name="MI250X",
        memory=128,
        architecture=AMDArchitecture.CDNA2,
        device_ids=(0x7408,),
    ),
    AMDGPUInfo(
        name="MI300A",
        memory=128,
        architecture=AMDArchitecture.CDNA3,
        device_ids=(0x74A0,),
    ),
    AMDGPUInfo(
        name="MI300X",
        memory=192,
        architecture=AMDArchitecture.CDNA3,
        device_ids=(0x74A1, 0x74A9, 0x74B5, 0x74BD),
    ),
    AMDGPUInfo(
        name="MI308X",
        memory=128,
        architecture=AMDArchitecture.CDNA3,
        device_ids=(0x74A2, 0x74B6),
    ),
    AMDGPUInfo(
        name="MI325X",
        memory=288,
        architecture=AMDArchitecture.CDNA3,
        device_ids=(0x74A5,),
    ),
]

KNOWN_TPUS: list[TPUInfo] = [TPUInfo(name=version, memory=0) for version in _TPU_VERSIONS]

KNOWN_INTEL_ACCELERATORS: list[IntelAcceleratorInfo] = [
    IntelAcceleratorInfo(name="Gaudi", memory=32),  # HL-205
    IntelAcceleratorInfo(name="Gaudi2", memory=96),  # HL-225
    IntelAcceleratorInfo(name="Gaudi3", memory=128),
]

KNOWN_TENSTORRENT_ACCELERATORS: list[TenstorrentAcceleratorInfo] = [
    TenstorrentAcceleratorInfo(name="n150", memory=12),
    TenstorrentAcceleratorInfo(name="n300", memory=24),
]

KNOWN_ACCELERATORS: list[
    Union[NvidiaGPUInfo, AMDGPUInfo, TPUInfo, IntelAcceleratorInfo, TenstorrentAcceleratorInfo]
] = (
    KNOWN_NVIDIA_GPUS
    + KNOWN_AMD_GPUS
    + KNOWN_TPUS
    + KNOWN_INTEL_ACCELERATORS
    + KNOWN_TENSTORRENT_ACCELERATORS
)



import logging
from typing import Any, Optional, cast

import requests
from requests import Response

from gpuhunt import QueryFilter, RawCatalogItem
from gpuhunt._internal.constraints import (
    find_accelerators,
    get_gpu_vendor,
    is_nvidia_superchip,
)
from gpuhunt._internal.models import AcceleratorVendor, CPUArchitecture
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)

API_URL = "https://api.vultr.com/v2"


class VultrProvider(AbstractProvider):
    NAME = "vultr"

    def get(
        self, query_filter: Optional[QueryFilter] = None, balance_resources: bool = True
    ) -> list[RawCatalogItem]:
        offers = fetch_offers()
        return sorted(offers, key=lambda i: i.price)


def fetch_offers() -> list[RawCatalogItem]:
    """Fetch plans with types:
    1. Cloud GPU (vcg),
    2. Bare Metal (vbm),
    3. and other CPU plans, including:
        Cloud Compute (vc2),
        High Frequency Compute (vhf),
        High Performance (vhp),
        All optimized Cloud Types (voc)"""
    bare_metal_plans_response = _make_request("GET", "/plans-metal?per_page=500")
    other_plans_response = _make_request("GET", "/plans?type=all&per_page=500")
    return convert_response_to_raw_catalog_items(bare_metal_plans_response, other_plans_response)


def convert_response_to_raw_catalog_items(
    bare_metal_plans_response: Response, other_plans_response: Response
) -> list[RawCatalogItem]:
    catalog_items = []

    bare_metal_plans = bare_metal_plans_response.json()["plans_metal"]
    other_plans = other_plans_response.json()["plans"]

    for plan in bare_metal_plans:
        for location in plan["locations"]:
            catalog_item = get_bare_metal_plans(plan, location)
            if catalog_item:
                catalog_items.append(catalog_item)

    for plan in other_plans:
        for location in plan["locations"]:
            catalog_item = get_instance_plans(plan, location)
            if catalog_item:
                catalog_items.append(catalog_item)

    return catalog_items


def get_bare_metal_plans(plan: dict, location: str) -> Optional[RawCatalogItem]:
    cpu_arch = CPUArchitecture.X86
    gpu_count, gpu_name, gpu_memory, gpu_vendor = 0, None, None, None
    if "gpu" in plan["id"]:
        if plan["id"] not in BARE_METAL_GPU_DETAILS:
            logger.warning("Skipping unknown GPU plan %s", plan["id"])
            return None
        gpu_count, gpu_name, gpu_memory = BARE_METAL_GPU_DETAILS[plan["id"]]
        if is_nvidia_superchip(gpu_name):
            cpu_arch = CPUArchitecture.ARM
        gpu_vendor = get_gpu_vendor(gpu_name)
        if gpu_vendor is None:
            logger.warning("Unknown GPU vendor for plan %s, skipping", plan["id"])
            return None
    return RawCatalogItem(
        instance_name=plan["id"],
        location=location,
        price=plan["hourly_cost"],
        cpu_arch=cpu_arch.value,
        cpu=plan["cpu_threads"],
        memory=plan["ram"] / 1024,
        gpu_count=gpu_count,
        gpu_name=gpu_name,
        gpu_memory=gpu_memory,
        gpu_vendor=gpu_vendor,
        spot=False,
        disk_size=plan["disk"],
    )


def get_instance_plans(plan: dict, location: str) -> Optional[RawCatalogItem]:
    cpu_arch = CPUArchitecture.X86
    plan_type = plan["type"]
    if plan_type in ["vc2", "vhf", "vhp", "voc"]:
        return RawCatalogItem(
            instance_name=plan["id"],
            location=location,
            price=plan["hourly_cost"],
            cpu_arch=cpu_arch.value,
            cpu=plan["vcpu_count"],
            memory=plan["ram"] / 1024,
            gpu_count=0,
            gpu_name=None,
            gpu_memory=None,
            gpu_vendor=None,
            spot=False,
            disk_size=plan["disk"],
        )
    elif plan_type == "vcg":
        gpu_type = cast(Optional[str], plan.get("gpu_type"))
        if not gpu_type:
            logger.warning("Missing gpu_type for plan %s, skipping", plan["id"])
            return None
        if "_" not in gpu_type:
            logger.warning(
                "Failed to parse gpu_type %s for plan %s, skipping", gpu_type, plan["id"]
            )
            return None
        gpu_name = gpu_type.split("_")[1]
        gpu_vendor = get_gpu_vendor(gpu_name)
        if not gpu_vendor:
            logger.warning(
                "Failed to detect GPU vendor %s for plan %s, skipping", gpu_type, plan["id"]
            )
            return None
        gpu_memory = get_gpu_memory(gpu_name)
        if not gpu_memory:
            logger.warning(
                "Failed to detect GPU memory %s for plan %s, skipping", gpu_type, plan["id"]
            )
            return None
        gpu_memory_total = cast(int, plan["gpu_vram_gb"])
        # For fractional GPU, gpu_count=1
        gpu_count = max(1, gpu_memory_total // gpu_memory)
        if is_nvidia_superchip(gpu_name):
            cpu_arch = CPUArchitecture.ARM
        return RawCatalogItem(
            instance_name=plan["id"],
            location=location,
            price=plan["hourly_cost"],
            cpu_arch=cpu_arch.value,
            cpu=plan["vcpu_count"],
            memory=plan["ram"] / 1024,
            gpu_count=gpu_count,
            gpu_name=gpu_name,
            gpu_memory=gpu_memory_total / gpu_count,
            gpu_vendor=gpu_vendor,
            spot=False,
            disk_size=plan["disk"],
        )
    return None


def get_gpu_memory(gpu_name: str) -> Optional[int]:
    if gpu_name.upper() == "A100":
        return 80  # VULTR A100 instances have 80GB
    if accelerators := find_accelerators(
        names=[gpu_name], vendors=[AcceleratorVendor.NVIDIA, AcceleratorVendor.AMD]
    ):
        return accelerators[0].memory
    logger.warning(f"Unknown GPU {gpu_name}")
    return None


def _make_request(method: str, path: str, data: Any = None) -> Response:
    response = requests.request(
        method=method,
        url=API_URL + path,
        json=data,
        timeout=30,
    )
    response.raise_for_status()
    return response


BARE_METAL_GPU_DETAILS = {
    "vbm-48c-1024gb-4-a100-gpu": (4, "A100", 80),
    "vbm-112c-2048gb-8-h100-gpu": (8, "H100", 80),
    "vbm-112c-2048gb-8-a100-gpu": (8, "A100", 80),
    "vbm-64c-2048gb-8-l40-gpu": (8, "L40S", 48),
    "vbm-72c-480gb-gh200-gpu": (1, "GH200", 96),
    "vbm-256c-2048gb-8-mi300x-gpu": (8, "MI300X", 192),
}



import dataclasses

import pytest

import gpuhunt._internal.catalog as internal_catalog
from gpuhunt import AcceleratorVendor, Catalog, CatalogItem, RawCatalogItem
from gpuhunt.providers.datacrunch import (
    DataCrunchProvider,
    InstanceType,
    generate_instances,
    get_gpu_name,
    transform_instance,
)


@pytest.fixture
def raw_instance_types() -> list[dict]:
    # datacrunch.instance_types.get()
    one_gpu = {
        "best_for": [
            "Gargantuan ML models",
            "Multi-GPU training",
            "FP64 HPC",
            "NVLINK",
        ],
        "cpu": {"description": "30 CPU", "number_of_cores": 30},
        "deploy_warning": "H100: Use Nvidia driver 535 or higher for best performance",
        "description": "Dedicated Hardware Instance",
        "gpu": {"description": "1x H100 SXM5 80GB", "number_of_gpus": 1},
        "gpu_memory": {"description": "80GB GPU RAM", "size_in_gigabytes": 80},
        "id": "c01dd00d-0000-480b-ae4e-d429115d055b",
        "instance_type": "1H100.80S.30V",
        "memory": {"description": "120GB RAM", "size_in_gigabytes": 120},
        "model": "H100 80GB",
        "name": "H100 SXM5 80GB",
        "p2p": "",
        "price_per_hour": "3.95",
        "spot_price": "1.70",
        "storage": {"description": "dynamic"},
    }

    two_gpu = {
        "best_for": ["Large ML models", "FP32 calculations", "Single-GPU training"],
        "cpu": {"description": "20 CPU", "number_of_cores": 20},
        "deploy_warning": None,
        "description": "Dedicated Hardware Instance",
        "gpu": {"description": "2x RTX A6000 48GB", "number_of_gpus": 2},
        "gpu_memory": {"description": "96GB GPU RAM", "size_in_gigabytes": 96},
        "id": "07cf5dc1-a5d2-4972-ae4e-d429115d055b",
        "instance_type": "2A6000.20V",
        "memory": {"description": "120GB RAM", "size_in_gigabytes": 120},
        "model": "RTX A6000",
        "name": "RTX A6000 48GB",
        "p2p": "",
        "price_per_hour": "1.98",
        "spot_price": "0.70",
        "storage": {"description": "dynamic"},
    }

    cpu_instance = {
        "best_for": ["Running services", "API server", "Data transfers"],
        "cpu": {"description": "120 CPU", "number_of_cores": 120},
        "deploy_warning": None,
        "description": "Dedicated Hardware Instance",
        "gpu": {"description": "", "number_of_gpus": 0},
        "gpu_memory": {"description": "", "size_in_gigabytes": 0},
        "id": "ccc00007-a5d2-4972-ae4e-d429115d055b",
        "instance_type": "CPU.120V.480G",
        "memory": {"description": "480GB RAM", "size_in_gigabytes": 480},
        "model": "CPU Node",
        "name": "AMD EPYC",
        "p2p": "",
        "price_per_hour": "3.00",
        "spot_price": "1.50",
        "storage": {"description": "dynamic"},
    }

    minimal = {
        "best_for": [
            "Small ML models",
            "Multi-GPU training",
            "FP64 calculations",
            "NVLINK",
        ],
        "cpu": {"description": "6 CPU", "number_of_cores": 6},
        "deploy_warning": None,
        "description": "Dedicated Hardware Instance",
        "gpu": {"description": "1x Tesla V100 16GB", "number_of_gpus": 1},
        "gpu_memory": {"description": "16GB GPU RAM", "size_in_gigabytes": 16},
        "id": "04cf5dc1-a5d2-4972-ae4e-d429115d055b",
        "instance_type": "1V100.6V",
        "memory": {"description": "23GB RAM", "size_in_gigabytes": 23},
        "model": "Tesla V100",
        "name": "Tesla V100 16GB",
        "p2p": "",
        "price_per_hour": "0.89",
        "spot_price": "0.25",
        "storage": {"description": "225GB NVME", "size_in_gigabytes": 225},
    }

    return [one_gpu, two_gpu, cpu_instance, minimal]


@pytest.fixture
def availabilities() -> list[dict]:
    # datacrunch.instances.get_availabilities(is_spot=True)
    data = [
        {
            "location_code": "FIN-01",
            "availabilities": [
                "1A100.22V",
                "1RTX6000ADA.10V",
                "8RTX6000ADA.80V",
                "2A6000.20V",
                "1V100.6V",
                "2V100.10V",
                "4V100.20V",
                "8V100.48V",
                "CPU.4V.16G",
                "CPU.8V.32G",
                "CPU.16V.64G",
                "CPU.32V.128G",
                "CPU.64V.256G",
                "CPU.96V.384G",
            ],
        },
        {
            "location_code": "ICE-01",
            "availabilities": [
                "CPU.4V.16G",
                "CPU.8V.32G",
                "CPU.16V.64G",
                "CPU.32V.128G",
                "CPU.64V.256G",
                "CPU.96V.384G",
                "CPU.120V.480G",
            ],
        },
    ]
    return data


@pytest.fixture
def locations():
    # datacrunch.locations.get()
    return [
        {"code": "FIN-01", "name": "Finland 1", "country_code": "FI"},
        {"code": "FIN-02", "name": "Finland 2", "country_code": "FI"},
        {"code": "ICE-01", "name": "Iceland 1", "country_code": "IS"},
    ]


def instance_types(raw_instance_type: dict) -> InstanceType:
    instance = InstanceType(
        id=raw_instance_type["id"],
        instance_type=raw_instance_type["instance_type"],
        price_per_hour=raw_instance_type["price_per_hour"],
        spot_price_per_hour=raw_instance_type["spot_price"],
        description=raw_instance_type["description"],
        cpu=raw_instance_type["cpu"],
        gpu=raw_instance_type["gpu"],
        memory=raw_instance_type["memory"],
        gpu_memory=raw_instance_type["gpu_memory"],
        storage=raw_instance_type["storage"],
    )
    return instance


def list_available_instances(raw_instance_types, locations):
    spots = (True, False)
    locations = [loc["loc"] for loc in locations]
    instances = [instance_types(raw_instance_types[0])]
    list_instances = generate_instances(spots, locations, instances)

    assert len(list_instances) == 4
    assert [i.price for i in list_instances if i.spot] == [1, 70] * 2
    assert [i.price for i in list_instances if not i.spot] == [3.95] * 2


def test_gpu_name(caplog):
    assert get_gpu_name("1x H100 SXM5 80GB") == "H100"
    assert get_gpu_name("") is None


def transform(raw_catalog_items: list[RawCatalogItem]) -> list[CatalogItem]:
    items = []
    for raw in raw_catalog_items:
        item = CatalogItem(provider="datacrunch", **dataclasses.asdict(raw))
        items.append(item)
    return items


def test_available_query(mocker, raw_instance_types):
    catalog = Catalog(balance_resources=False, auto_reload=False)

    instance_type = instance_types(raw_instance_types[0])

    mocker.patch("datacrunch.DataCrunchClient.__init__", return_value=None)
    datacrunch = DataCrunchProvider("EXAMPLE", "EXAMPLE")
    datacrunch._get_instance_types = mocker.Mock(return_value=[instance_type])
    datacrunch._get_locations = mocker.Mock(return_value=[{"code": "FIN-01"}])

    internal_catalog.ONLINE_PROVIDERS = ["datacrunch"]
    internal_catalog.OFFLINE_PROVIDERS = []

    catalog.add_provider(datacrunch)
    query_result = catalog.query(provider=["datacrunch"])

    assert len(query_result) == 2

    expected_spot = CatalogItem(
        instance_name="1H100.80S.30V",
        location="FIN-01",
        price=1.7,
        cpu=30,
        memory=120.0,
        gpu_vendor=AcceleratorVendor.NVIDIA,
        gpu_count=1,
        gpu_name="H100",
        gpu_memory=80.0,
        spot=True,
        provider="datacrunch",
        disk_size=None,
    )
    expected_non_spot = CatalogItem(
        instance_name="1H100.80S.30V",
        location="FIN-01",
        price=3.95,
        cpu=30,
        memory=120.0,
        gpu_vendor=AcceleratorVendor.NVIDIA,
        gpu_count=1,
        gpu_name="H100",
        gpu_memory=80.0,
        spot=False,
        provider="datacrunch",
        disk_size=None,
    )
    assert [r for r in query_result if r.spot] == [expected_spot]
    assert [r for r in query_result if not r.spot] == [expected_non_spot]


def test_available_query_with_instance(mocker, raw_instance_types):
    catalog = Catalog(balance_resources=False, auto_reload=False)

    instance_type = instance_types(raw_instance_types[-1])
    print(instance_type)

    mocker.patch("datacrunch.DataCrunchClient.__init__", return_value=None)
    datacrunch = DataCrunchProvider("EXAMPLE", "EXAMPLE")
    datacrunch._get_instance_types = mocker.Mock(return_value=[instance_type])
    datacrunch._get_locations = mocker.Mock(return_value=[{"code": "FIN-01"}])

    internal_catalog.ONLINE_PROVIDERS = ["datacrunch"]
    internal_catalog.OFFLINE_PROVIDERS = []

    catalog.add_provider(datacrunch)
    query_result = catalog.query(provider=["datacrunch"])

    print(query_result)

    assert len(query_result) == 2

    expected_spot = CatalogItem(
        instance_name="1V100.6V",
        location="FIN-01",
        price=0.25,
        cpu=6,
        memory=23.0,
        gpu_vendor=AcceleratorVendor.NVIDIA,
        gpu_count=1,
        gpu_name="V100",
        gpu_memory=16.0,
        spot=True,
        provider="datacrunch",
        disk_size=None,
    )
    expected_non_spot = CatalogItem(
        instance_name="1V100.6V",
        location="FIN-01",
        price=0.89,
        cpu=6,
        memory=23.0,
        gpu_vendor=AcceleratorVendor.NVIDIA,
        gpu_count=1,
        gpu_name="V100",
        gpu_memory=16.0,
        spot=False,
        provider="datacrunch",
        disk_size=None,
    )

    assert [r for r in query_result if r.spot] == [expected_spot]
    assert [r for r in query_result if not r.spot] == [expected_non_spot]


def test_transform_instance(raw_instance_types):
    location = "ICE-01"
    is_spot = True
    item = transform_instance(instance_types(raw_instance_types[1]), is_spot, location)

    expected = RawCatalogItem(
        instance_name="2A6000.20V",
        location="ICE-01",
        price=0.7,
        cpu=20,
        memory=120,
        gpu_vendor=AcceleratorVendor.NVIDIA.value,
        gpu_count=2,
        gpu_name="A6000",
        gpu_memory=96 / 2,
        spot=True,
        disk_size=None,
    )

    assert RawCatalogItem.from_dict(item) == expected


def test_cpu_instance(raw_instance_types):
    location = "ICE-01"
    is_spot = False
    item = transform_instance(instance_types(raw_instance_types[2]), is_spot, location)

    expected = RawCatalogItem(
        instance_name="CPU.120V.480G",
        location="ICE-01",
        price=3,
        cpu=120,
        memory=480,
        gpu_vendor=None,
        gpu_count=0,
        gpu_name=None,
        gpu_memory=None,
        spot=False,
        disk_size=None,
    )

    assert RawCatalogItem.from_dict(item) == expected


def test_order(mocker, raw_instance_types):
    catalog = Catalog(balance_resources=False, auto_reload=False)

    types = map(instance_types, raw_instance_types)

    mocker.patch("datacrunch.DataCrunchClient.__init__", return_value=None)
    datacrunch = DataCrunchProvider("EXAMPLE", "EXAMPLE")
    datacrunch._get_instance_types = mocker.Mock(return_value=list(types))
    datacrunch._get_locations = mocker.Mock(return_value=[{"code": "FIN-01"}])

    internal_catalog.ONLINE_PROVIDERS = ["datacrunch"]
    internal_catalog.OFFLINE_PROVIDERS = []

    catalog.add_provider(datacrunch)
    query_result = catalog.query(provider=["datacrunch"])

    assert len(query_result) == 8

    assert [r.price for r in query_result] == sorted(r.price for r in query_result)



import csv
import dataclasses
import heapq
import io
import logging
import os
import time
import urllib.request
import zipfile
from collections.abc import Container
from concurrent.futures import ThreadPoolExecutor, wait
from pathlib import Path
from typing import Optional, Union

import gpuhunt._internal.constraints as constraints
from gpuhunt._internal.models import AcceleratorVendor, CatalogItem, CPUArchitecture, QueryFilter
from gpuhunt._internal.utils import parse_compute_capability
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)

VERSION_URL = "https://dstack-gpu-pricing.s3.eu-west-1.amazonaws.com/v2/version"
CATALOG_URL = "https://dstack-gpu-pricing.s3.eu-west-1.amazonaws.com/v2/{version}/catalog.zip"
OFFLINE_PROVIDERS = [
    "aws",
    "azure",
    "datacrunch",
    "gcp",
    "lambdalabs",
    "nebius",
    "oci",
    "runpod",
    "cloudrift",
]
ONLINE_PROVIDERS = ["cudo", "digitalocean", "hotaisle", "tensordock", "vastai", "vultr"]
RELOAD_INTERVAL = 15 * 60  # 15 minutes


class Catalog:
    def __init__(self, balance_resources: bool = True, auto_reload: bool = True):
        """
        Args:
            balance_resources: increase min resources to better match the chosen GPU
            auto_reload: if `True`, the catalog will be automatically loaded from the S3 bucket every 4 hours
        """
        self.catalog = None
        self.loaded_at = None
        self.providers: list[AbstractProvider] = []
        self.balance_resources = balance_resources
        self.auto_reload = auto_reload

    def query(
        self,
        *,
        provider: Optional[Union[str, list[str]]] = None,
        cpu_arch: Optional[Union[CPUArchitecture, str]] = None,
        min_cpu: Optional[int] = None,
        max_cpu: Optional[int] = None,
        min_memory: Optional[float] = None,
        max_memory: Optional[float] = None,
        min_gpu_count: Optional[int] = None,
        max_gpu_count: Optional[int] = None,
        gpu_vendor: Optional[Union[AcceleratorVendor, str]] = None,
        gpu_name: Optional[Union[str, list[str]]] = None,
        min_gpu_memory: Optional[float] = None,
        max_gpu_memory: Optional[float] = None,
        min_total_gpu_memory: Optional[float] = None,
        max_total_gpu_memory: Optional[float] = None,
        min_disk_size: Optional[int] = None,
        max_disk_size: Optional[int] = None,
        min_price: Optional[float] = None,
        max_price: Optional[float] = None,
        min_compute_capability: Optional[Union[str, tuple[int, int]]] = None,
        max_compute_capability: Optional[Union[str, tuple[int, int]]] = None,
        spot: Optional[bool] = None,
        allowed_flags: Optional[Container[str]] = None,
    ) -> list[CatalogItem]:
        """
        Query the catalog for matching offers

        Args:
            provider: name of the provider to filter by. If not specified, all providers will be used
            cpu_arch: CPU architecture to filter by. If not specified, all architectures will be used
            min_cpu: minimum number of CPUs
            max_cpu: maximum number of CPUs
            min_memory: minimum amount of RAM in GB
            max_memory: maximum amount of RAM in GB
            min_gpu_count: minimum number of GPUs
            max_gpu_count: maximum number of GPUs
            gpu_vendor: accelerator vendor to filter by. If not specified, all vendors will be used
            gpu_name: name of the GPU to filter by. If not specified, all GPUs will be used
            min_gpu_memory: minimum amount of GPU VRAM in GB for each GPU
            max_gpu_memory: maximum amount of GPU VRAM in GB for each GPU
            min_total_gpu_memory: minimum amount of GPU VRAM in GB for all GPUs combined
            max_total_gpu_memory: maximum amount of GPU VRAM in GB for all GPUs combined
            min_disk_size: minimum disk size in GB
            max_disk_size: maximum disk size in GB
            min_price: minimum price per hour in USD
            max_price: maximum price per hour in USD
            min_compute_capability: minimum compute capability of the GPU
            max_compute_capability: maximum compute capability of the GPU
            spot: if `False`, only ondemand offers will be returned. If `True`, only spot offers will be returned
            allowed_flags: only offers with all flags allowed will be returned. `None` allows all flags

        Returns:
            list of matching offers
        """
        if self.auto_reload and (
            self.loaded_at is None or time.monotonic() - self.loaded_at > RELOAD_INTERVAL
        ):
            self.load()

        query_filter = QueryFilter(
            provider=[provider] if isinstance(provider, str) else provider,
            cpu_arch=CPUArchitecture.cast(cpu_arch) if cpu_arch else None,
            min_cpu=min_cpu,
            max_cpu=max_cpu,
            min_memory=min_memory,
            max_memory=max_memory,
            min_gpu_count=min_gpu_count,
            max_gpu_count=max_gpu_count,
            gpu_vendor=AcceleratorVendor.cast(gpu_vendor) if gpu_vendor else None,
            gpu_name=[gpu_name] if isinstance(gpu_name, str) else gpu_name,
            min_gpu_memory=min_gpu_memory,
            max_gpu_memory=max_gpu_memory,
            min_total_gpu_memory=min_total_gpu_memory,
            max_total_gpu_memory=max_total_gpu_memory,
            min_disk_size=min_disk_size,
            max_disk_size=max_disk_size,
            min_price=min_price,
            max_price=max_price,
            min_compute_capability=parse_compute_capability(min_compute_capability),
            max_compute_capability=parse_compute_capability(max_compute_capability),
            spot=spot,
            allowed_flags=allowed_flags,
        )

        if query_filter.provider is not None:
            # validate providers
            for p in query_filter.provider:
                if p.lower() not in OFFLINE_PROVIDERS + ONLINE_PROVIDERS:
                    raise ValueError(f"Unknown provider: {p}")
        else:
            query_filter.provider = OFFLINE_PROVIDERS + list(
                set(p.NAME for p in self.providers if p.NAME in ONLINE_PROVIDERS)
            )

        # fetch providers
        with ThreadPoolExecutor(max_workers=8) as executor:
            futures = []

            for provider_name in ONLINE_PROVIDERS:
                if provider_name in map(str.lower, query_filter.provider):
                    futures.append(
                        executor.submit(
                            self._get_online_provider_items,
                            provider_name,
                            query_filter,
                        )
                    )

            for provider_name in OFFLINE_PROVIDERS:
                if provider_name in map(str.lower, query_filter.provider):
                    futures.append(
                        executor.submit(
                            self._get_offline_provider_items,
                            provider_name,
                            query_filter,
                        )
                    )

            completed, _ = wait(futures)
            # The merge preserves provider-specific order, picking the cheapest offer at each step.
            # The final list is not strictly sorted by the price.
            items = list(heapq.merge(*[f.result() for f in completed], key=lambda i: i.price))
        return items

    def load(self, version: Optional[str] = None):
        """
        Fetch the catalog from the S3 bucket

        Args:
            version: specific version of the catalog to download. If not specified, the latest version will be used
        """
        catalog_url = os.getenv("GPUHUNT_CATALOG_URL")
        if catalog_url is None:
            if version is None:
                version = self.get_latest_version()
            catalog_url = CATALOG_URL.format(version=version)
        logger.debug("Downloading catalog %s...", version)
        with urllib.request.urlopen(catalog_url) as f:
            self.loaded_at = time.monotonic()
            self.catalog = io.BytesIO(f.read())

    @staticmethod
    def get_latest_version() -> str:
        """
        Get the latest version of the catalog from the S3 bucket
        """
        with urllib.request.urlopen(VERSION_URL) as f:
            return f.read().decode("utf-8").strip()

    def add_provider(self, provider: AbstractProvider):
        """
        Add provider for querying offers

        Args:
            provider: provider to add
        """
        self.providers.append(provider)

    def _get_offline_provider_items(
        self, provider_name: str, query_filter: QueryFilter
    ) -> list[CatalogItem]:
        logger.debug("Loading items for offline provider %s", provider_name)
        items = []
        # Set this env var to use a local catalog instead of the s3 catalog
        catalog_dir = os.getenv("GPUHUNT_CATALOG_DIR")
        if catalog_dir is not None:
            with open(Path(catalog_dir) / f"{provider_name}.csv", "rb") as csv_file:
                reader = csv.DictReader(io.TextIOWrapper(csv_file, "utf-8"))
                for row in reader:
                    item = CatalogItem.from_dict(row, provider=provider_name)
                    if constraints.matches(item, query_filter):
                        items.append(item)
        else:
            if self.catalog is None:
                logger.warning("Catalog not loaded")
                return items
            with zipfile.ZipFile(self.catalog) as zip_file:
                with zip_file.open(f"{provider_name}.csv", "r") as csv_file:
                    reader = csv.DictReader(io.TextIOWrapper(csv_file, "utf-8"))
                    for row in reader:
                        item = CatalogItem.from_dict(row, provider=provider_name)
                        if constraints.matches(item, query_filter):
                            items.append(item)
        return items

    def _get_online_provider_items(
        self, provider_name: str, query_filter: QueryFilter
    ) -> list[CatalogItem]:
        logger.debug("Loading items for online provider %s", provider_name)
        items = []
        found = False
        for provider in self.providers:
            if provider.NAME != provider_name:
                continue
            found = True
            for i in provider.get(
                query_filter=query_filter, balance_resources=self.balance_resources
            ):
                item = CatalogItem(provider=provider_name, **dataclasses.asdict(i))
                if constraints.matches(item, query_filter):
                    items.append(item)
        if not found:
            raise ValueError(f"Provider is not loaded: {provider_name}")
        return items
